import flax.linen as nn
import jax.numpy as jnp


class MyModel(nn.Module):
    output_dim: int

    @nn.compact
    def __call__(self, x, train: bool = False, **kwargs) -> jnp.ndarray:
        x = x.reshape(x.shape[0], -1)
        x = nn.Dense(2, name="l1")(x)
        x = nn.Dropout(rate=0.9)(x, deterministic=not train)
        x = nn.Dense(self.output_dim, name="l2")(x)
        return x



from typing import Dict, Optional, Tuple, Union

import flax.linen as nn
import jax.numpy as jnp
from flax.core import FrozenDict
from flax.training.checkpoints import PyTree
from jax import random
from jax._src.prng import PRNGKeyArray

from fortuna.model.model_manager.base import ModelManager
from fortuna.typing import Array, Mutable, Params


class ClassificationModelManager(ModelManager):
    def __init__(self, model: nn.Module):
        r"""
        Classification model manager class. It orchestrates the forward pass of the model in the probabilistic model.

        Parameters
        ----------
        model : nn.Module
            A model describing the deterministic relation between inputs and outputs. The outputs must correspond to
            the logits of a softmax probability vector. The output dimension must be the same as the number of classes.
            Let :math:`x` be input variables and :math:`w` the random model parameters. Then the model is described by
            a function :math:`f(w, x)`, where each component of :math:`f` corresponds to one of the classes.
        """
        self.model = model

    def apply(
        self,
        params: Params,
        inputs: Array,
        mutable: Optional[Mutable] = None,
        train: bool = False,
        rng: Optional[PRNGKeyArray] = None,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, PyTree]]:
        if mutable is None:
            mutable = False
        variables = params["model"].unfreeze()

        # setup dropout key
        if rng is not None:
            rng, dropout_rng = random.split(rng, 2)
            rngs = {"dropout": dropout_rng}
        else:
            rngs = None

        if mutable:
            mutable_variables = mutable["model"].unfreeze()
            variables.update(mutable_variables)
            mutable = list(mutable_variables.keys())
        if train and mutable:
            outputs, mutable = self.model.apply(
                variables, inputs, mutable=mutable, train=train, rngs=rngs
            )
            return outputs, {"mutable": FrozenDict({"model": mutable})}
        else:
            return self.model.apply(
                variables, inputs, train=train, mutable=False, rngs=rngs
            )

    def init(
        self, input_shape: Tuple[int, ...], rng: Optional[PRNGKeyArray] = None, **kwargs
    ) -> Dict[str, FrozenDict]:
        if rng is None:
            rng = self.rng.get()
        rng, params_key, dropout_key = random.split(rng, 3)
        rngs = {"params": params_key, "dropout": dropout_key}
        return dict(
            model=self.model.init(rngs, jnp.zeros((1,) + input_shape), **kwargs)
        )



import enum

from fortuna.calibration.state import CalibState
from fortuna.training.train_state import TrainState


class NameToTrainState(enum.Enum):
    vars()[TrainState.__name__] = TrainState
    vars()[CalibState.__name__] = CalibState



from __future__ import annotations

import logging
from typing import Dict, Optional

import jax.numpy as jnp
from flax.core import FrozenDict
from jax import hessian, lax, vjp
from jax._src.prng import PRNGKeyArray
from jax.flatten_util import ravel_pytree
from jax.tree_util import tree_map

from fortuna.data.loader import DataLoader
from fortuna.prob_model.fit_config import FitConfig
from fortuna.prob_model.joint.base import Joint
from fortuna.prob_model.joint.state import JointState
from fortuna.prob_model.posterior.base import Posterior
from fortuna.prob_model.posterior.laplace import LAPLACE_NAME
from fortuna.prob_model.posterior.laplace.laplace_approximator import \
    LaplacePosteriorApproximator
from fortuna.prob_model.posterior.laplace.laplace_state import LaplaceState
from fortuna.prob_model.posterior.map.map_approximator import \
    MAPPosteriorApproximator
from fortuna.prob_model.posterior.map.map_posterior import MAPPosterior
from fortuna.prob_model.posterior.map.map_state import MAPState
from fortuna.prob_model.posterior.posterior_state_repository import \
    PosteriorStateRepository
from fortuna.prob_model.prior.gaussian import (DiagonalGaussianPrior,
                                               IsotropicGaussianPrior)
from fortuna.typing import CalibMutable, CalibParams, Mutable, Params, Status
from fortuna.utils.nested_dicts import nested_get, nested_set, nested_unpair
from fortuna.utils.random import generate_random_normal_like_tree


class LaplacePosterior(Posterior):
    def __init__(
        self, joint: Joint, posterior_approximator: LaplacePosteriorApproximator,
    ):
        """
        Laplace approximation posterior class.

        Parameters
        ----------
        joint: Joint
            A joint distribution object.
        posterior_approximator: LaplacePosteriorApproximator
            A Laplace posterior approximator.
        """
        super().__init__(joint=joint, posterior_approximator=posterior_approximator)
        if type(joint.prior) not in [DiagonalGaussianPrior, IsotropicGaussianPrior]:
            raise ValueError(
                """The Laplace posterior_approximation is not supported for this model. The prior distribution must be one of the 
                following choices: {}.""".format(
                    [DiagonalGaussianPrior, IsotropicGaussianPrior]
                )
            )

    def __str__(self):
        return LAPLACE_NAME

    def _gnn_approx(
        self,
        params: Params,
        train_data_loader: DataLoader,
        mutable: Optional[Mutable] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        factorization: str = "diagonal",
        verbose: bool = True,
    ) -> Params:
        """
        Estimate a standard deviation for each parameter using a diagonal Generalized Gauss-Newton Hessian
        approximation.

        Parameters
        ----------
        params : Params
            The random parameters of the probabilistic model.
        train_data_loader: DataLoader
            A training data loader.
        mutable: Optional[Mutable]
            Mutable objects.
        calib_params : Optional[CalibParams]
            The calibration parameters of the probabilistic model.
        calib_mutable : Optional[CalibMutable] = None
            The calibration mutable objects used to evaluate the calibrators.
        factorization: str = "diagonal"
            Factorization of the GGN approximation. Currently, only "diagonal" is supported.
        verbose: bool
            Whether to log the training progress.

        Returns
        -------
        Params
            An estimate of the posterior standard deviation for each random parameter.
        """
        rav, unravel = ravel_pytree(
            tuple(
                [
                    nested_get(params, keys)
                    for keys in self.posterior_approximator.which_params
                ]
            )
            if self.posterior_approximator.which_params
            else params
        )

        def get_params_from_rav(_rav):
            unrav = unravel(_rav)
            return (
                FrozenDict(
                    nested_set(
                        params.unfreeze(),
                        self.posterior_approximator.which_params,
                        unrav,
                    )
                )
                if self.posterior_approximator.which_params
                else unrav
            )

        def apply_calib_model_manager(_params, _batch_inputs):
            outputs = self.joint.likelihood.model_manager.apply(
                _params, _batch_inputs, mutable=mutable, train=False
            )
            outputs = self.joint.likelihood.output_calib_manager.apply(
                params=calib_params["output_calibrator"]
                if calib_params is not None
                else None,
                mutable=calib_mutable["output_calibrator"]
                if calib_mutable is not None
                else None,
                outputs=outputs,
            )
            return outputs

        apply_fn = lambda _rav, x: apply_calib_model_manager(
            get_params_from_rav(_rav), x
        ).squeeze(0)
        vjp_fn = lambda x: vjp(lambda _rav: apply_fn(_rav, x), rav)[1]

        def eig_hess_fn(vars):
            hess = hessian(
                lambda __o: self.joint.likelihood.prob_output_layer.log_prob(
                    __o, vars[1]
                )
            )(vars[0])
            return jnp.linalg.eigh(hess)

        def compute_hess_batch(_batch_inputs, _batch_targets):
            lam, z = lax.map(
                eig_hess_fn,
                (apply_calib_model_manager(params, batch_inputs), batch_targets),
            )
            ztj = lax.map(
                lambda v: lax.map(vjp_fn(v[0]), v[1].T), (batch_inputs[:, None], z)
            )[0]
            if factorization == "diagonal":
                return -jnp.sum(lam[:, :, None] * ztj ** 2, (0, 1))
            raise ValueError(
                f"`factorization={factorization}` not recognized. Currently, only "
                f"`factorization='diagonal'` is supported."
            )

        h = jnp.exp(-self.joint.prior.log_var)
        for i, (batch_inputs, batch_targets) in enumerate(train_data_loader):
            if verbose:
                logging.info(f"Hessian approximation for batch {i + 1}.")
            h += compute_hess_batch(batch_inputs, batch_targets)

        return unravel(1 / jnp.sqrt(h))

    def fit(
        self,
        train_data_loader: DataLoader,
        val_data_loader: Optional[DataLoader] = None,
        fit_config: FitConfig = FitConfig(),
        map_fit_config: Optional[FitConfig] = None,
        **kwargs,
    ) -> Dict[str, Status]:
        if (
            fit_config.checkpointer.dump_state is True
            and not fit_config.checkpointer.save_checkpoint_dir
        ):
            raise ValueError(
                "`save_checkpoint_dir` must be passed when `dump_state` is set to True."
            )

        status = {}

        if not fit_config.checkpointer.restore_checkpoint_path:
            map_posterior = MAPPosterior(
                self.joint, posterior_approximator=MAPPosteriorApproximator()
            )
            map_posterior.rng = self.rng
            logging.info("Do a preliminary run of MAP.")
            status["map"] = map_posterior.fit(
                rng=self.rng.get(),
                train_data_loader=train_data_loader,
                val_data_loader=val_data_loader,
                fit_config=map_fit_config
                if map_fit_config is not None
                else FitConfig(),
                **kwargs,
            )
            logging.info("Preliminary run with MAP completed.")
            state = map_posterior.state.get()
        else:
            state = self.restore_checkpoint(
                restore_checkpoint_path=fit_config.checkpointer.restore_checkpoint_path
            )

        if type(state) == MAPState:
            logging.info("Run the Laplace approximation.")
            std_params = self._gnn_approx(
                state.params,
                train_data_loader,
                mutable=state.mutable,
                calib_params=state.calib_params,
                calib_mutable=state.calib_mutable,
                verbose=fit_config.monitor.verbose,
            )
            state = LaplaceState.convert_from_map_state(
                map_state=state,
                std=std_params,
                which_params=self.posterior_approximator.which_params,
            )

        if fit_config.checkpointer.save_checkpoint_dir:
            self.save_checkpoint(
                state,
                save_checkpoint_dir=fit_config.checkpointer.save_checkpoint_dir,
                keep=fit_config.checkpointer.keep_top_n_checkpoints,
                force_save=True,
            )

        self.state = PosteriorStateRepository(
            fit_config.checkpointer.save_checkpoint_dir
            if fit_config.checkpointer.dump_state is True
            else None
        )
        self.state.put(state, keep=fit_config.checkpointer.keep_top_n_checkpoints)
        logging.info("Fit completed.")
        return status

    def sample(self, rng: Optional[PRNGKeyArray] = None, **kwargs) -> JointState:
        if rng is None:
            rng = self.rng.get()
        state = self.state.get()

        if self.posterior_approximator.which_params:
            mean, std = nested_unpair(
                state.params.unfreeze(),
                self.posterior_approximator.which_params,
                ("mean", "std"),
            )

            noise = generate_random_normal_like_tree(rng, std)
            params = nested_set(
                mean,
                self.posterior_approximator.which_params,
                tuple(
                    [
                        tree_map(
                            lambda m, s, e: m + s * e,
                            nested_get(mean, keys),
                            nested_get(std, keys),
                            nested_get(noise, keys),
                        )
                        for keys in self.posterior_approximator.which_params
                    ]
                ),
            )
            for k, v in params.items():
                params[k] = FrozenDict(v)
            state = state.replace(params=FrozenDict(params))
        else:
            mean, std = dict(), dict()
            for k, v in state.params.items():
                mean[k] = FrozenDict({"params": v["params"]["mean"]})
                std[k] = FrozenDict({"params": v["params"]["std"]})

            state = state.replace(
                params=FrozenDict(
                    tree_map(
                        lambda m, s, e: m + s * e,
                        mean,
                        std,
                        generate_random_normal_like_tree(rng, std),
                    )
                )
            )

        return JointState(
            params=state.params,
            mutable=state.mutable,
            calib_params=state.calib_params,
            calib_mutable=state.calib_mutable,
        )



import enum

from fortuna.prob_model.posterior.deep_ensemble import DEEP_ENSEMBLE_NAME
from fortuna.prob_model.posterior.deep_ensemble.deep_ensemble_posterior import \
    DeepEnsemblePosterior
from fortuna.prob_model.posterior.laplace import LAPLACE_NAME
from fortuna.prob_model.posterior.laplace.laplace_posterior import \
    LaplacePosterior
from fortuna.prob_model.posterior.map import MAP_NAME
from fortuna.prob_model.posterior.map.map_posterior import MAPPosterior
from fortuna.prob_model.posterior.normalizing_flow.advi import ADVI_NAME
from fortuna.prob_model.posterior.normalizing_flow.advi.advi_posterior import \
    ADVIPosterior
from fortuna.prob_model.posterior.swag import SWAG_NAME
from fortuna.prob_model.posterior.swag.swag_posterior import SWAGPosterior


class PosteriorApproximations(enum.Enum):
    """Map approximator name to posterior posterior_approximation."""

    vars()[MAP_NAME] = MAPPosterior
    vars()[ADVI_NAME] = ADVIPosterior
    vars()[DEEP_ENSEMBLE_NAME] = DeepEnsemblePosterior
    vars()[LAPLACE_NAME] = LaplacePosterior
    vars()[SWAG_NAME] = SWAGPosterior



from fortuna.prob_model.fit_config.base import FitConfig
from fortuna.prob_model.fit_config.checkpointer import FitCheckpointer
from fortuna.prob_model.fit_config.monitor import FitMonitor
from fortuna.prob_model.fit_config.optimizer import FitOptimizer
from fortuna.prob_model.fit_config.processor import FitProcessor



from typing import Callable, List, Optional, Tuple, Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.collections import Collection
from matplotlib.patches import Circle, RegularPolygon
from matplotlib.path import Path
from matplotlib.projections import register_projection
from matplotlib.projections.polar import PolarAxes
from matplotlib.spines import Spine
from matplotlib.transforms import Affine2D

from fortuna.data.loader import InputsLoader
from fortuna.prob_model.regression import ProbRegressor
from fortuna.typing import Array, Batch


def plot_2d_classification_predictions_and_uncertainty(
    inputs: Optional[np.ndarray] = None,
    preds: Optional[np.ndarray] = None,
    grid_uncertainty: Optional[np.ndarray] = None,
    uncertainty_fn: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    grid: Optional[np.ndarray] = None,
    xx1: Optional[np.ndarray] = None,
    xx2: Optional[np.ndarray] = None,
    ax: Optional[matplotlib.axes.Axes] = None,
    figsize: tuple = (5, 5),
    title: Optional[str] = None,
    uncertainty_cmap: str = "Blues",
    preds_color: Tuple[str, str] = ("C2", "C3"),
    base_inputs_color: str = "black",
    marker_size: int = 10,
    legend: bool = True,
    uncertainty_label: str = "uncertainty",
    fontsize: int = 12,
    colorbar: bool = False,
    remove_ticks: bool = False,
    show: bool = False,
    return_pcolor: bool = False,
    **save_options,
) -> Optional[Collection]:
    """
    Plot inputs, predictions and uncertainty.

    :param inputs: Optional[np.ndarray]
        Input data. This must be a two-dimensional array, with different inputs over the first axis. The number of
        dimensions over the second axis must be 2.
    :param preds: Optional[np.ndarray]
        Predictions. This must be a one-dimensional array of zeros and ones, with different inputs over the first
        axis.
    :param grid_uncertainty: Optional[np.ndarray]
        Scalar uncertainty value for each element of the grid. This must have the same dimensionality as `grid`.
    :param uncertainty_fn: Optional[Callable[[np.ndarray], np.ndarray]]
        A function taking as argument a grid of inputs, and returning a scalar uncertainty value for each element
        in the grid.
    :param grid: Optional[np.ndarray]
        The grid to display the plot over.
    :param xx1: Optional[np.ndarray]
        Mesh over the first axis.
    :param xx2: Optional[np.ndarray]
        Mesh over the second axis.
    :param ax: Optional[matplotlib.axes.Axes]
        Axis matplotlib object.
    :param figsize: tuple
        Figure size.
    :param title: Optional[str]
        Figure title.
    :param uncertainty_cmap: str
        Uncertainty color map.
    :param preds_color: Tuple[str, str]
        Color of predictions. The first element of the tuple is associated to the prediction 1; the second to 0.
    :param base_inputs_color: str
        Color of inputs if `preds` is not available.
    :param marker_size: int
        Size of scattered inputs.
    :param legend: bool
        Whether to include a legend.
    :param uncertainty_label: str
        Uncertainty label in the legend.
    :param fontsize: int
        Font size in the legend.
    :param colorbar: bool
        Whether to include a color bar.
    :param remove_ticks: bool
        Whether to remove ticks from axes.
    :param save_options: dict
        Options to save the file with `matplotlib.pyplot.savefig`. If no option is given, the file is not saved.
    :param show: bool
        Whether to show the plot.
    :param return_pcolor: bool
        Whether to return the output of `matplotlib.pyplot.pcolor`.

    :return Optional[Collection]
        If `return_pcolor`, it returns the output of `matplotlib.pyplot.pcolor`.
    """
    if (xx1 is None or xx2 is None) and (inputs is None):
        raise ValueError(
            """If `inputs` is not passed, please pass either both `xx1` and `xx2`, or `grid`.
        """
        )
    if grid_uncertainty is None and uncertainty_fn is None:
        raise ValueError(
            """ One between `grid_uncertainty` and `uncertainty_fn` must be passed. If both are passed,
        `uncertainty_fn` is ignored."""
        )
    if grid is None and uncertainty_fn is None:
        raise ValueError(
            """If `grid` is not passed, then `uncertainty_fn` must be passed."""
        )
    if grid is not None and (uncertainty_fn is None and grid_uncertainty is None):
        raise ValueError(
            """If `grid` is not passed, then at least one of `uncertainty_fn` and `grid_uncertainty`
        must be passed."""
        )
    if grid_uncertainty is not None and grid is None:
        raise ValueError(
            """If `grid_uncertainty` is passed, then `grid` must also be passed."""
        )
    if grid is None:
        if xx1 is None or xx2 is None:
            mins, maxs = np.min(inputs, 0), np.max(inputs, 0)
            if xx1 is None:
                xx1 = np.linspace(
                    mins[0] - 1.25 * (maxs[0] - mins[0]),
                    maxs[0] + 1.25 * (maxs[0] - mins[0]),
                    100 if xx2 is None else len(xx2),
                )
            if xx2 is None:
                xx2 = np.linspace(
                    mins[1] - 1.25 * (maxs[1] - mins[1]),
                    maxs[1] + 1.25 * (maxs[1] - mins[1]),
                    100 if xx1 is None else len(xx1),
                )
        grid = np.array([[_xx1, _xx2] for _xx1 in xx1 for _xx2 in xx2])
    if grid_uncertainty is None:
        grid_uncertainty = uncertainty_fn(grid)
    size1 = len(np.unique(grid[:, 0])) if xx1 is None else len(xx1)
    size2 = len(np.unique(grid[:, 1])) if xx2 is None else len(xx2)
    grid_uncertainty = grid_uncertainty.reshape(size1, size2)
    grid = grid.reshape(size1, size2, 2)
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=figsize)
    if title:
        ax.set_title(title, fontsize=fontsize + 2)
    im = ax.pcolor(
        grid[:, :, 0],
        grid[:, :, 1],
        grid_uncertainty,
        cmap=uncertainty_cmap,
        label=uncertainty_label,
    )
    if inputs is not None:
        ax.scatter(
            inputs[:, 0],
            inputs[:, 1],
            s=marker_size,
            c=[preds_color[0] if i == 1 else preds_color[1] for i in preds]
            if preds is not None
            else base_inputs_color,
        )
    if colorbar:
        plt.colorbar(im, ax=ax.ravel().tolist() if hasattr(ax, "ravel") else ax)
    if legend:
        ax.legend(fontsize=fontsize)
    if remove_ticks:
        ax.xticks([], [])
        ax.yticks([], [])
    if len(save_options) > 0:
        plt.savefig(**save_options)
    if show:
        plt.show()

    if return_pcolor:
        return im


def plot_reliability_diagram(
    accs: Union[Array, List[Array]],
    confs: Union[Array, List[Array]],
    ax: Optional[matplotlib.axes.Axes] = None,
    labels: Optional[List[str]] = None,
    figsize: tuple = (6, 4),
    legend_loc: Optional[str] = None,
    fontsize: int = 10,
    title: Optional[str] = None,
    ylim: Optional[Tuple[float, float]] = None,
    show: bool = False,
    **save_options,
) -> None:
    """
    Plot a reliability diagram.

    Parameters
    ----------
    accs: Union[Array, List[Array, ...]]
        An accuracy score for each bin. A list of multiple accuracy scores is also accepted.
    confs: Union[Array, List[Array, ...]]
        A confidence score for each bin. A list of multiple confidence scores is also accepted.
    labels: Optional[List[str, ...]]
        Labels used in the legend presumably describing the method used to generate accuracy and confidence scores. A
        list of labels, corresponding to lists of accuracies and confidences, is also accepted.
    ax: Optional[matplotlib.axes.Axes]
        Axis matplotlib object.
    figsize: Optional[tuple]
        Figure size.
    legend_loc: Optional[str]
        Legend location, as in :code:`matplotlib`.
        See `here <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.legend.html>`__
    fontsize: Optional[int]
        Font size.
    title: Optional[str] = None
        Plot title.
    ylim: Optional[Tuple[float, float]]
        Bottom and top limits on the y-axis.
    show: bool
        Whether to show the plot.
    save_options: dict
        Options to save the file with `matplotlib.pyplot.savefig`. If no option is given, the file is not saved.
    """
    if figsize is None:
        figsize = (6, 4)
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=figsize)
    if type(confs) != list:
        confs = [confs]
    if type(accs) != list:
        accs = [accs]
    if labels:
        if type(labels) != list:
            labels = [labels]
    if len(accs) != len(confs):
        raise ValueError("`accs` and `confs` must contain the same number of entries.")
    if labels:
        if len(accs) != len(labels):
            raise ValueError(
                "`accs` and `labels` must contain the same number of entries."
            )
    ax.grid()
    ax.plot([0, 1], [0, 0], color="grey", linestyle="--", alpha=0.3)
    for i, (a, c) in enumerate(zip(accs, confs)):
        ax.plot(
            c, c - a, marker=".", linestyle="-", label=labels[i] if labels else None
        )
    if labels:
        ax.legend(fontsize=fontsize, loc=legend_loc if legend_loc else None)
    ax.set_xlabel("confidence", fontsize=fontsize)
    ax.set_ylabel("confidence - accuracy", fontsize=fontsize)
    if title:
        ax.set_title(title, fontsize=fontsize + 2)
    if ylim:
        ax.set_ylim(ylim)
    if len(save_options) > 0:
        plt.savefig(**save_options)
    if show:
        plt.show()
    if ax:
        return ax


def plot_1d_regression_preds_and_std(
    prob_model: ProbRegressor,
    mesh: np.ndarray,
    data: Optional[Batch] = None,
    ax: Optional[matplotlib.axes.Axes] = None,
    n_std: float = 2.0,
    title: Optional[str] = None,
    figsize: Tuple = (5, 2),
    fontsize: int = 12,
    marker_size_data: int = 1,
    legend: bool = True,
    show: bool = False,
    **saving_options,
) -> Optional[Collection]:
    mesh_loader = InputsLoader.from_array_inputs(mesh)
    if not ax:
        fig, ax = plt.subplots(1, 1, figsize=figsize)
    if title:
        ax.set_title(title, fontsize=fontsize + 2)

    mean = prob_model.predictive.mean(mesh_loader)
    std = prob_model.predictive.std(mesh_loader)
    if data:
        ax.scatter(*data, s=marker_size_data, color="C0", label="data")
    ax.plot(mesh, mean, color="C1", label="mean")
    ax.fill_between(
        mesh,
        (mean - n_std * std).squeeze(1),
        (mean + n_std * std).squeeze(1),
        alpha=0.3,
        color="C0",
        label=f"+/- {n_std} std",
    )
    if legend:
        ax.legend(fontsize=fontsize - 2)
    plt.tight_layout()
    if show:
        plt.show()
    if len(saving_options) > 0:
        plt.savefig(**saving_options)
    return ax


def radar_factory(num_vars, frame="circle"):
    """
    Create a radar chart with `num_vars` axes.

    This function creates a RadarAxes projection and registers it.

    Parameters
    ----------
    num_vars : int
        Number of variables for radar chart.
    frame : {'circle', 'polygon'}
        Shape of frame surrounding axes.

    """
    # calculate evenly-spaced axis angles
    theta = np.linspace(0, 2 * np.pi, num_vars, endpoint=False)

    class RadarTransform(PolarAxes.PolarTransform):
        def transform_path_non_affine(self, path):
            # Paths with non-unit interpolation steps correspond to gridlines,
            # in which case we force interpolation (to defeat PolarTransform's
            # autoconversion to circular arcs).
            if path._interpolation_steps > 1:
                path = path.interpolated(num_vars)
            return Path(self.transform(path.vertices), path.codes)

    class RadarAxes(PolarAxes):

        name = "radar"
        PolarTransform = RadarTransform

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            # rotate plot such that the first axis is at the top
            self.set_theta_zero_location("N")

        def fill(self, *args, closed=True, **kwargs):
            """Override fill so that line is closed by default"""
            return super().fill(closed=closed, *args, **kwargs)

        def plot(self, *args, **kwargs):
            """Override plot so that line is closed by default"""
            lines = super().plot(*args, **kwargs)
            for line in lines:
                self._close_line(line)

        def _close_line(self, line):
            x, y = line.get_data()
            # FIXME: markers at x[0], y[0] get doubled-up
            if x[0] != x[-1]:
                x = np.append(x, x[0])
                y = np.append(y, y[0])
                line.set_data(x, y)

        def set_varlabels(self, labels):
            self.set_thetagrids(np.degrees(theta), labels)

        def _gen_axes_patch(self):
            # The Axes patch must be centered at (0.5, 0.5) and of radius 0.5
            # in axes coordinates.
            if frame == "circle":
                return Circle((0.5, 0.5), 0.5)
            elif frame == "polygon":
                return RegularPolygon((0.5, 0.5), num_vars, radius=0.5, edgecolor="k")
            else:
                raise ValueError("Unknown value for 'frame': %s" % frame)

        def _gen_axes_spines(self):
            if frame == "circle":
                return super()._gen_axes_spines()
            elif frame == "polygon":
                # spine_type must be 'left'/'right'/'top'/'bottom'/'circle'.
                spine = Spine(
                    axes=self,
                    spine_type="circle",
                    path=Path.unit_regular_polygon(num_vars),
                )
                # unit_regular_polygon gives a polygon of radius 1 centered at
                # (0, 0) but we want a polygon of radius 0.5 centered at (0.5,
                # 0.5) in axes coordinates.
                spine.set_transform(
                    Affine2D().scale(0.5).translate(0.5, 0.5) + self.transAxes
                )
                return {"polar": spine}
            else:
                raise ValueError("Unknown value for 'frame': %s" % frame)

    register_projection(RadarAxes)
    return theta



from __future__ import annotations

import logging
from typing import Optional, Tuple

import jax.numpy as jnp
from flax.core import FrozenDict
from jax._src.prng import PRNGKeyArray
from jax.flatten_util import ravel_pytree

from fortuna.data.loader import DataLoader, InputsLoader
from fortuna.distribution.gaussian import DiagGaussian
from fortuna.prob_model.fit_config import FitConfig
from fortuna.prob_model.joint.base import Joint
from fortuna.prob_model.joint.state import JointState
from fortuna.prob_model.posterior.base import Posterior
from fortuna.prob_model.posterior.normalizing_flow.advi import ADVI_NAME
from fortuna.prob_model.posterior.normalizing_flow.advi.advi_approximator import \
    ADVIPosteriorApproximator
from fortuna.prob_model.posterior.normalizing_flow.advi.advi_architecture import \
    ADVIArchitecture
from fortuna.prob_model.posterior.normalizing_flow.advi.advi_state import \
    ADVIState
from fortuna.prob_model.posterior.normalizing_flow.advi.advi_trainer import \
    ADVITrainer
from fortuna.prob_model.posterior.posterior_state_repository import \
    PosteriorStateRepository
from fortuna.training.trainer import JittedMixin, MultiDeviceMixin
from fortuna.typing import Array, Status
from fortuna.utils.device import select_trainer_given_devices


class JittedADVITrainer(JittedMixin, ADVITrainer):
    pass


class MultiDeviceADVITrainer(MultiDeviceMixin, ADVITrainer):
    pass


class ADVIPosterior(Posterior):
    def __init__(
        self, joint: Joint, posterior_approximator: ADVIPosteriorApproximator,
    ):
        """
        Automatic Differentiation Variational Inference (ADVI) approximate posterior class.

        Parameters
        ----------
        joint: Joint
            A joint distribution object.
        posterior_approximator: ADVI
            An ADVI posterior approximator.
        """
        super().__init__(joint=joint, posterior_approximator=posterior_approximator)

    def __str__(self):
        return ADVI_NAME

    def fit(
        self,
        train_data_loader: DataLoader,
        val_data_loader: Optional[DataLoader] = None,
        fit_config: FitConfig = FitConfig(),
        **kwargs,
    ) -> Status:
        if (
            fit_config.checkpointer.dump_state is True
            and not fit_config.checkpointer.save_checkpoint_dir
        ):
            raise ValueError(
                "`save_checkpoint_dir` must be passed when `dump_state` is set to True."
            )

        init_prob_model_state, n_train_data, n_val_data = self._init(
            train_data_loader, val_data_loader
        )

        rav, self.unravel = ravel_pytree(init_prob_model_state.params)
        size_rav = len(rav)
        self.base = DiagGaussian(
            mean=jnp.zeros(size_rav),
            std=self.posterior_approximator.std_base * jnp.ones(size_rav),
        )
        self.architecture = ADVIArchitecture(
            size_rav, std_init_params=self.posterior_approximator.std_init_params
        )

        trainer_cls = select_trainer_given_devices(
            devices=fit_config.processor.devices,
            BaseTrainer=ADVITrainer,
            JittedTrainer=JittedADVITrainer,
            MultiDeviceTrainer=MultiDeviceADVITrainer,
            disable_jit=fit_config.processor.disable_jit,
        )

        trainer = trainer_cls(
            predict_fn=self.joint.likelihood.prob_output_layer.predict,
            save_checkpoint_dir=fit_config.checkpointer.save_checkpoint_dir,
            save_every_n_steps=fit_config.checkpointer.save_every_n_steps,
            keep_top_n_checkpoints=fit_config.checkpointer.keep_top_n_checkpoints,
            disable_training_metrics_computation=fit_config.monitor.disable_training_metrics_computation,
            eval_every_n_epochs=fit_config.monitor.eval_every_n_epochs,
            early_stopping_monitor=fit_config.monitor.early_stopping_monitor,
            early_stopping_min_delta=fit_config.monitor.early_stopping_min_delta,
            early_stopping_patience=fit_config.monitor.early_stopping_patience,
            base=self.base,
            architecture=self.architecture,
        )

        state = None
        if fit_config.checkpointer.restore_checkpoint_path:
            state = self.restore_checkpoint(
                restore_checkpoint_path=fit_config.checkpointer.restore_checkpoint_path,
                optimizer=fit_config.optimizer.method,
            )

        if type(state) != ADVIState:
            state = ADVIState.init(
                FrozenDict(
                    zip(
                        ("mean", "logvar"),
                        trainer.init_params(
                            self.rng.get(),
                            mean=ravel_pytree(
                                getattr(state, "params", init_prob_model_state.params)
                            )[0],
                        ),
                    )
                ),
                getattr(state, "mutable", init_prob_model_state.mutable),
                fit_config.optimizer.method,
                getattr(state, "calib_params", init_prob_model_state.calib_params),
                getattr(state, "calib_mutable", init_prob_model_state.calib_mutable),
            )
        logging.info("Run ADVI.")
        state, status = trainer.train(
            rng=self.rng.get(),
            state=state,
            fun=self.joint._batched_log_joint_prob,
            training_dataloader=train_data_loader,
            training_dataset_size=n_train_data,
            n_epochs=fit_config.optimizer.n_epochs,
            metrics=fit_config.monitor.metrics,
            validation_dataloader=val_data_loader,
            validation_dataset_size=n_val_data,
            verbose=fit_config.monitor.verbose,
            unravel=self.unravel,
            n_samples=self.posterior_approximator.n_loss_samples,
        )
        self.state = PosteriorStateRepository(
            fit_config.checkpointer.save_checkpoint_dir
            if fit_config.checkpointer.dump_state is True
            else None
        )
        self.state.put(state, keep=fit_config.checkpointer.keep_top_n_checkpoints)
        logging.info("Fit completed.")
        return status

    def sample(
        self,
        rng: Optional[PRNGKeyArray] = None,
        input_shape: Optional[Tuple[int, ...]] = None,
        inputs_loader: Optional[InputsLoader] = None,
        inputs: Optional[Array] = None,
        **kwargs,
    ) -> JointState:
        """
        Sample from the posterior distribution. Either `input_shape` or `_inputs_loader` must be passed.

        Parameters
        ----------
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        input_shape: Optional[Tuple[int, ...]]
            Shape of a single input.
        inputs_loader: Optional[InputsLoader]
            Input data loader. If `input_shape` is passed, then `inputs` and `inputs_loader` are ignored.
        inputs: Optional[Array]
            Input variables.

        Returns
        -------
        JointState
            A sample from the posterior distribution.
        """
        if rng is None:
            rng = self.rng.get()
        state = self.state.get()
        state = state.replace(params=tuple(state.params.values()))
        n_params = len(state.params[0])
        if not hasattr(self, "base"):
            self.base = DiagGaussian(
                mean=jnp.zeros(n_params),
                std=self.posterior_approximator.std_base * jnp.ones(n_params),
            )
        if not hasattr(self, "architecture"):
            self.architecture = ADVIArchitecture(
                n_params, std_init_params=self.posterior_approximator.std_init_params
            )

        if not hasattr(self, "unravel"):
            if input_shape is None:
                if inputs is not None:
                    input_shape = inputs.shape[1:]
                else:
                    if inputs_loader is None:
                        raise ValueError(
                            "Either `input_shape` or `inputs_loader` or `inputs` must be passed."
                        )
                    for x in inputs_loader:
                        input_shape = x.shape[1:]
                        break
            model_manager_state = self.joint.likelihood.model_manager.init(input_shape)
            self.unravel = ravel_pytree(model_manager_state.params)[1]
        sample_params = self.unravel(
            self.architecture.forward(state.params, self.base.sample(rng))[0][0]
        )

        return JointState(
            params=FrozenDict(sample_params),
            mutable=state.mutable,
            calib_params=state.calib_params,
            calib_mutable=state.calib_mutable,
        )



from __future__ import annotations

from typing import Any

import jax.numpy as jnp
from flax.training import train_state

from fortuna.utils.strings import convert_string_to_jnp_array


class TrainState(train_state.TrainState):
    encoded_name: jnp.ndarray = convert_string_to_jnp_array("TrainState")

    @classmethod
    def init(cls, *args, **kwargs) -> Any:
        pass

    @classmethod
    def init_from_dict(cls, *args, **kwargs) -> TrainState:
        pass



import abc
from typing import Optional, Tuple, Union

from jax._src.prng import PRNGKeyArray

from fortuna.data.loader import DataLoader
from fortuna.prob_model.fit_config import FitConfig
from fortuna.prob_model.joint.base import Joint
from fortuna.prob_model.joint.state import JointState
from fortuna.prob_model.posterior.posterior_mixin import \
    WithPosteriorCheckpointingMixin
from fortuna.prob_model.posterior.posterior_state_repository import \
    PosteriorStateRepository
from fortuna.typing import Path, Status
from fortuna.utils.random import WithRNG


class PosteriorApproximator(abc.ABC):
    """
    A posterior approximator abstract class.
    """

    @abc.abstractmethod
    def __str__(self):
        pass


class Posterior(WithRNG, WithPosteriorCheckpointingMixin):
    state = None

    def __init__(self, joint: Joint, posterior_approximator: PosteriorApproximator):
        r"""
        Posterior distribution class. This refers to :math:`p(w|\mathcal{D}, \phi)`, where :math:`w` are the random
        model parameters, :math:`\mathcal{D}` is a training data set and :math:`\phi` are calibration parameters.

        Parameters
        ----------
        joint: Joint
            A joint distribution object.
        posterior_approximator: PosteriorApproximator
            A posterior approximator.
        """
        super().__init__()
        self.joint = joint
        self.posterior_approximator = posterior_approximator

    def _init(
        self,
        train_data_loader: DataLoader,
        val_data_loader: Optional[DataLoader] = None,
    ) -> Tuple[JointState, int, Union[int, None]]:
        n_train_data = 0
        for i, (batch_inputs, batch_targets) in enumerate(train_data_loader):
            n_train_data += batch_targets.shape[0]
            if i == 0:
                input_shape = batch_inputs.shape[1:]

        n_val_data = None
        if val_data_loader is not None:
            n_val_data = 0
            for batch in val_data_loader:
                n_val_data += batch[1].shape[0]

        return self.joint.init(input_shape), n_train_data, n_val_data

    @abc.abstractmethod
    def fit(
        self,
        train_data_loader: DataLoader,
        val_data_loader: Optional[DataLoader] = None,
        fit_config: FitConfig = FitConfig(),
        **kwargs,
    ) -> Status:
        """
        Fit the posterior distribution. A posterior state will be internally stored.

        Parameters
        ----------
        train_data_loader: DataLoader
            Training data loader.
        val_data_loader: Optional[DataLoader]
            Validation data loader.
        fit_config: FitConfig
            A configuration object.

        Returns
        -------
        Status
            A status including metrics describing the fitting process.
        """
        pass

    @abc.abstractmethod
    def sample(self, rng: Optional[PRNGKeyArray] = None, *args, **kwargs) -> JointState:
        """
        Sample from the posterior distribution.

        Parameters
        ----------
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.

        Returns
        -------
        JointState
            A sample from the posterior distribution.
        """
        pass

    def load_state(self, checkpoint_path: Path) -> None:
        """
        Load the state of the posterior distribution from a checkpoint path. The checkpoint must be
        compatible with the current probabilistic model.

        Parameters
        ----------
        checkpoint_path: Path
            Path to checkpoint file or directory to restore.
        """
        try:
            self.restore_checkpoint(checkpoint_path)
        except ValueError:
            raise ValueError(
                f"No checkpoint was found in `checkpoint_path={checkpoint_path}`."
            )
        self.state = PosteriorStateRepository(checkpoint_dir=checkpoint_path)

    def save_state(
        self, checkpoint_path: Path, keep_top_n_checkpoints: int = 1
    ) -> None:
        """
        Save the state of the posterior distribution to a checkpoint directory.

        Parameters
        ----------
        checkpoint_path: Path
            Path to checkpoint file or directory to restore.
        keep_top_n_checkpoints: int
            Number of past checkpoint files to keep.
        """
        if self.state is None:
            raise ValueError(
                """No state available. You must first either fit the posterior distribution, or load a 
            saved checkpoint."""
            )
        return self.state.put(
            self.state.get(),
            checkpoint_path=checkpoint_path,
            keep=keep_top_n_checkpoints,
        )



import numpy as np

from fortuna.data.loader import DataLoader


def check_data_loader_is_not_random(data_loader: DataLoader) -> None:
    flag = False
    for (x1, y1), (x2, y2) in zip(data_loader, data_loader):
        if not np.alltrue(x1 == x2) or not np.alltrue(y1 == y2):
            flag = True
            break
    if flag:
        raise ValueError(
            """The data loader randomizes at every iteration. To perform this method, please provide a data loader that 
            generates the same sequence of data when called multiple times."""
        )



from typing import Dict, Optional, Tuple, Union

import flax.linen as nn
import jax.numpy as jnp
from flax.core import FrozenDict
from flax.training.checkpoints import PyTree
from jax import random
from jax._src.prng import PRNGKeyArray

from fortuna.model.model_manager.base import ModelManager
from fortuna.typing import Array, Mutable, Params


class RegressionModelManager(ModelManager):
    def __init__(self, model: nn.Module, likelihood_log_variance_model: nn.Module):
        r"""
        Regression model manager class. It orchestrates the forward pass of the model in the probabilistic model.

        Parameters
        ----------
        model : nn.Module
            A model describing the deterministic relation between inputs and outputs. It characterizes the mean model
            of the likelihood function. The outputs must belong to the same space as the target variables.
            Let :math:`x` be input variables and :math:`w` the random model parameters. Then the model is described by
            a function :math:`\mu(w, x)`.
        likelihood_log_variance_model: nn.Module
            A model characterizing the log-variance of a Gaussian likelihood function. The outputs must belong to the
            same space as the target variables. Let :math:`x` be input variables and :math:`w` the random model
            parameters. Then the model is described by a function :math:`\log\sigma^2(w, x)`.
        """
        self.model = model
        self.likelihood_log_variance_model = likelihood_log_variance_model

    def apply(
        self,
        params: Params,
        inputs: Array,
        mutable: Optional[Mutable] = None,
        train: bool = False,
        rng: Optional[PRNGKeyArray] = None,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, PyTree]]:
        if mutable is None:
            mutable = False
        model_variables = params["model"].unfreeze()
        lik_log_var_variables = params["lik_log_var"].unfreeze()

        # setup dropout key
        if rng is not None:
            rng, model_dropout_key, lik_log_var_dropout_key = random.split(rng, 3)
            model_rngs = {"dropout": model_dropout_key}
            lik_log_var_rngs = {"dropout": lik_log_var_dropout_key}
        else:
            model_rngs = None
            lik_log_var_rngs = None

        if mutable:
            try:
                model_mutable_variables = mutable["model"].unfreeze()
                model_variables.update(model_mutable_variables)
                model_mutable = list(model_mutable_variables.keys())
            except KeyError:
                model_mutable = False
            try:
                lik_log_var_mutable_variables = mutable["lik_log_var"].unfreeze()
                lik_log_var_variables.update(lik_log_var_mutable_variables)
                lik_log_var_mutable = list(lik_log_var_mutable_variables.keys())
            except KeyError:
                lik_log_var_mutable = False
        else:
            model_mutable = False
            lik_log_var_mutable = False

        aux = {}
        if train and model_mutable:
            model_outputs, model_mutable = self.model.apply(
                model_variables,
                inputs,
                train=train,
                mutable=model_mutable,
                rngs=model_rngs,
            )
            aux.update({"model": model_mutable})
        else:
            model_outputs = self.model.apply(
                model_variables, inputs, train=train, mutable=False, rngs=model_rngs
            )
        if train and lik_log_var_mutable:
            (
                lik_log_var_outputs,
                lik_log_var_mutable,
            ) = self.likelihood_log_variance_model.apply(
                lik_log_var_variables,
                train=train,
                mutable=lik_log_var_mutable,
                rngs=lik_log_var_rngs,
            )
            aux.update({"likelihood_log_variance_model": lik_log_var_mutable})
        else:
            lik_log_var_outputs = self.model.apply(
                lik_log_var_variables,
                inputs,
                train=train,
                mutable=False,
                rngs=lik_log_var_rngs,
            )

        outputs = jnp.concatenate((model_outputs, lik_log_var_outputs), axis=-1)
        if model_outputs.shape[-1] != lik_log_var_outputs.shape[-1]:
            raise ValueError(
                f"""The output dimensions of `model` and `likelihood_log_variance_model must be the same. 
            However, {model_outputs.shape[-1]} and {lik_log_var_outputs.shape[-1]} were found, respectively."""
            )
        if len(aux) > 0:
            return outputs, {"mutable": FrozenDict(aux)}
        return outputs

    def init(
        self, input_shape: Tuple, rng: Optional[PRNGKeyArray] = None, **kwargs
    ) -> Dict[str, FrozenDict]:
        if rng is None:
            rng = self.rng.get()
        (
            rng,
            model_params_key,
            model_dropout_key,
            lik_log_var_params_key,
            lik_log_var_dropout_key,
        ) = random.split(rng, 5)
        model_rngs = {"params": model_params_key, "dropout": model_dropout_key}
        lik_log_var_rngs = {
            "params": lik_log_var_params_key,
            "dropout": lik_log_var_params_key,
        }
        return dict(
            model=self.model.init(model_rngs, jnp.zeros((1,) + input_shape), **kwargs),
            lik_log_var=self.model.init(
                lik_log_var_rngs, jnp.zeros((1,) + input_shape), **kwargs
            ),
        )



from jax import numpy as jnp
from jax.tree_util import tree_flatten


def tree_stack(trees):
    """Takes a list of trees and stacks every corresponding leaf.
    For example, given two trees ((a, b), c) and ((a', b'), c'), returns
    ((stack(a, a'), stack(b, b')), stack(c, c')).
    Useful for turning a list of objects into something you can feed to a
    vmapped function.
    """
    leaves_list = []
    treedef_list = []
    for tree in trees:
        leaves, treedef = tree_flatten(tree)
        leaves_list.append(leaves)
        treedef_list.append(treedef)

    grouped_leaves = zip(*leaves_list)
    result_leaves = [jnp.stack(l) for l in grouped_leaves]
    return treedef_list[0].unflatten(result_leaves)


def tree_unstack(tree):
    """Takes a tree and turns it into a list of trees. Inverse of tree_stack.
    For example, given a tree ((a, b), c), where a, b, and c all have first
    dimension k, will make k trees
    [((a[0], b[0]), c[0]), ..., ((a[k], b[k]), c[k])]
    Useful for turning the output of a vmapped function into normal objects.
    """
    leaves, treedef = tree_flatten(tree)
    n_trees = leaves[0].shape[0]
    new_leaves = [[] for _ in range(n_trees)]
    for leaf in leaves:
        for i in range(n_trees):
            new_leaves[i].append(leaf[i])
    new_trees = [treedef.unflatten(l) for l in new_leaves]
    return new_trees


from fortuna.calib_config.base import CalibConfig
from fortuna.calib_config.checkpointer import CalibCheckpointer
from fortuna.calib_config.monitor import CalibMonitor
from fortuna.calib_config.optimizer import CalibOptimizer
from fortuna.calib_config.processor import CalibProcessor



from fortuna.prob_model.posterior.deep_ensemble.deep_ensemble_approximator import \
    DeepEnsemblePosteriorApproximator
from fortuna.prob_model.posterior.laplace.laplace_approximator import \
    LaplacePosteriorApproximator
from fortuna.prob_model.posterior.map.map_posterior import \
    MAPPosteriorApproximator
from fortuna.prob_model.posterior.normalizing_flow.advi.advi_approximator import \
    ADVIPosteriorApproximator
from fortuna.prob_model.posterior.swag.swag_approximator import \
    SWAGPosteriorApproximator



from __future__ import annotations

import abc
from typing import Any, Callable, Dict, Optional, Tuple, Union

import jax.numpy as jnp
import numpy as np
from flax.core import FrozenDict
from jax import random, vmap
from jax._src.prng import PRNGKeyArray
from jax.tree_util import tree_map
from optax._src.base import PyTree

from fortuna.distribution.base import Distribution
from fortuna.prob_model.posterior.posterior_trainer import PosteriorTrainerABC
from fortuna.prob_model.posterior.state import PosteriorState
from fortuna.typing import Array, Batch, CalibMutable, CalibParams, Params


class NormalizingFlowTrainer(PosteriorTrainerABC):
    @abc.abstractmethod
    def __str__(self):
        pass

    def __init__(self, *, base: Distribution, architecture: object, **kwargs):
        """
        This is a Normalizing Flow. It is able to transform a `base` distribution into a target one via an invertible
        `architecture`.

        :param base: object
            Base distribution. It must include the following methods:
                - sample(rng: jnp.array, n_samples: int) -> jnp.array
                    It draws `n_samples` samples from the distribution.
                - log_joint_prob(x: jnp.array) -> float
                    It evaluates the log-probability density function at `x`.
            Common distribution are already available in `prob.distribution`.
        :param architecture: object
            Invertible architecture. It must include the following methods:
                - forward(params: jnp.array, u: jnp.array) ->  jnp.array
                    It applies on `u` the transformation parametrized by `params`.
                - backward(params: jnp.array, v: jnp.array) ->  jnp.array
                    It applies on `v` the inverse transformation parametrized by `params`.
                - init_params(rng: jnp.array)
                    It initializes the parameters of the transformation.
        """
        super(NormalizingFlowTrainer, self).__init__(**kwargs)
        # base distribution
        self.sample_base = base.sample
        self.base_log_joint_prob = base.log_joint_prob

        # architecture
        self.forward = architecture.forward
        self.backward = architecture.backward
        self.init_params = architecture.init_params

    def sample_forward(self, rng: jnp.array, params: any, n_samples: int) -> tuple:
        """
        Sample from the push-forward distribution.

        :param rng: jnp.array
            Random number generator.
        :param params: any
            Transformation parameters.
        :param n_samples:
            Number of samples to draw.

        :return: tuple
            forward_samples: jnp.array
                Samples.
            ldj: jnp.array
                Log-determinant of the Jacobians.
        """
        return self.forward(params, self.sample_base(rng, n_samples))

    def training_loss_step(
        self,
        fun: Callable[[Any], Union[float, Tuple[float, dict]]],
        params: Params,
        batch: Tuple[Union[jnp.ndarray, np.ndarray], Union[jnp.ndarray, np.ndarray]],
        mutable: FrozenDict[str, FrozenDict],
        rng: jnp.ndarray,
        n_data: int,
        unravel: Optional[Callable[[any], PyTree]] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        kwargs: FrozenDict[str, Any] = FrozenDict(),
    ) -> Tuple[jnp.ndarray, Dict[str, Any]]:
        return_aux = ["outputs"]
        if mutable is not None:
            return_aux += ["mutable"]
        rng, key = random.split(rng)
        v, ldj = self.sample_forward(key, tuple(params.values()), kwargs["n_samples"])
        logp, aux = vmap(
            lambda _params: fun(
                unravel(_params),
                batch,
                n_data=n_data,
                mutable=mutable,
                return_aux=return_aux,
                train=True,
                rng=rng,
                calib_params=calib_params,
                calib_mutable=calib_mutable,
            )
        )(v)
        return (
            -(jnp.mean(logp) + jnp.mean(ldj)),
            {
                "outputs": aux.get("outputs"),
                "mutable": tree_map(lambda x: x[-1], aux.get("mutable")),
                "logging_kwargs": dict(),
            },
        )

    def training_step_end(
        self,
        current_epoch: int,
        state: PosteriorState,
        aux: Dict[str, Any],
        batch: Batch,
        metrics: Optional[Tuple[Callable[[jnp.ndarray, Array], float], ...]] = None,
        kwargs: FrozenDict[str, Any] = FrozenDict(),
    ) -> Dict[str, jnp.ndarray]:
        """
        Perform training batch metrics_names computation and save checkpoint if needed.

        :param current_epoch: int
            Current epoch.
        :param state: TrainState
            The training state.
        :param aux: Dict[str, Any]
            The dictionary obtained from a call to `training_loss_step()`. It must have a key name `outputs` which
            contains the model's prediction for the given `batch`.
        :param batch: Batch
            The input data and the targets.
        :param metrics: Optional[Tuple[Callable[[jnp.ndarray, Array], float], ...]]
            A tuple of metrics.
        :param kwargs: FrozenDict[str, Any]
            Any other extra argument. They have to be explicitly passed within a dictionary and cannot be provided as
            named arguments due to `jax.jit`.

        :return: Dict[str, jnp.ndarray]
            A dictionary containing `metrics_names` values on the given input batch.
        """
        if (
            self.save_checkpoint_dir
            and self.save_every_n_steps
            and current_epoch % self.save_every_n_steps
        ):
            self.save_checkpoint(
                state, self.save_checkpoint_dir, keep=self.keep_top_n_checkpoints
            )
        training_losses_and_metrics = {"loss": aux["loss"]}

        if aux["logging_kwargs"] is not None:
            for k, v in aux["logging_kwargs"].items():
                training_losses_and_metrics[k] = v

        if not self.disable_training_metrics_computation and metrics is not None:
            preds = self.predict_fn(aux["outputs"])
            if self.multi_device:
                training_batch_metrics = vmap(
                    lambda p: self.compute_metrics(
                        p,
                        batch[1].reshape(
                            (batch[1].shape[0] * batch[1].shape[1],)
                            + batch[1].shape[2:]
                        ),
                        metrics,
                    )
                )(
                    preds.reshape(
                        (preds.shape[1], preds.shape[0] * preds.shape[2],)
                        + preds.shape[3:]
                    )
                )
            else:
                training_batch_metrics = vmap(
                    lambda p: self.compute_metrics(p, batch[1], metrics)
                )(preds)
            training_batch_metrics = tree_map(
                lambda m: m.mean(), training_batch_metrics
            )
            for k, v in training_batch_metrics.items():
                training_losses_and_metrics[k] = v
        return training_losses_and_metrics

    def validation_step(
        self,
        state: PosteriorState,
        batch: Batch,
        fun: Callable[[Any], Union[float, Tuple[float, dict]]],
        rng: PRNGKeyArray,
        n_data: int,
        metrics: Optional[Tuple[Callable[[jnp.ndarray, Array], float]], ...] = None,
        unravel: Optional[Callable[[any], PyTree]] = None,
        kwargs: FrozenDict[str, Any] = FrozenDict(),
    ) -> Dict[str, jnp.ndarray]:
        rng, key = random.split(rng)
        v, ldj = self.sample_forward(
            key, tuple(state.params.values()), kwargs["n_samples"]
        )
        logp, aux = vmap(
            lambda _params: fun(
                unravel(_params),
                batch,
                n_data=n_data,
                mutable=state.mutable,
                return_aux=["outputs"],
                train=False,
                rng=rng,
                calib_params=state.calib_params,
                calib_mutable=state.calib_mutable,
            )
        )(v)
        loss = -(jnp.mean(logp) + jnp.mean(ldj))
        if metrics is not None:
            preds = self.predict_fn(aux["outputs"])
            val_metrics = vmap(lambda p: self.compute_metrics(p, batch[1], metrics))(
                preds
            )
            val_metrics = tree_map(lambda m: m.mean(), val_metrics)
            return {
                "val_loss": loss,
                **{f"val_{m}": v for m, v in val_metrics.items()},
            }
        return dict(val_loss=loss)



from fortuna.model.lenet import LeNet5
from fortuna.model.mlp import MLP
from fortuna.model.resnet import (ResNet18, ResNet34, ResNet50, ResNet101,
                                  ResNet152, ResNet200)
from fortuna.model.wideresnet import WideResNet28_10



import pathlib
from typing import Dict, Tuple, Union

import jax.numpy as jnp
import numpy as np
from flax.core import FrozenDict
from optax._src.base import GradientTransformation, PyTree

Params = FrozenDict[str, FrozenDict[str, PyTree]]
Mutable = FrozenDict[str, FrozenDict[str, PyTree]]
CalibParams = FrozenDict[str, PyTree]
CalibMutable = FrozenDict[str, PyTree]
OptaxOptimizer = GradientTransformation
Array = Union[jnp.ndarray, np.ndarray]
Status = Dict[str, Array]
Path = Union[str, pathlib.Path]
Batch = Tuple[Array, Array]



import abc
import collections
import logging
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import jax
import jax.numpy as jnp
from flax import jax_utils
from flax.training.common_utils import stack_forest
from jax import lax, random, value_and_grad
from jax._src.prng import PRNGKeyArray
from jax.tree_util import tree_map
from tqdm import trange
from tqdm.std import tqdm as TqdmDecorator

from fortuna.calibration.state import CalibState
from fortuna.data.loader import DataLoader, TargetsLoader
from fortuna.training.mixin import (InputValidatorMixin,
                                    WithCheckpointingMixin,
                                    WithEarlyStoppingMixin)
from fortuna.typing import Array, CalibMutable, CalibParams, Path, Status
from fortuna.utils.builtins import HashableMixin


class CalibModelCalibrator(
    HashableMixin,
    WithCheckpointingMixin,
    WithEarlyStoppingMixin,
    InputValidatorMixin,
    metaclass=abc.ABCMeta,
):
    def __init__(
        self,
        *args,
        calib_outputs: Array,
        calib_targets: Array,
        predict_fn: Callable[[jnp.ndarray], jnp.ndarray],
        uncertainty_fn: Callable[[jnp.ndarray], jnp.ndarray],
        val_outputs: Array,
        val_targets: Array,
        save_checkpoint_dir: Optional[Path] = None,
        save_every_n_steps: Optional[int] = None,
        keep_top_n_checkpoints: int = 2,
        disable_training_metrics_computation: bool = False,
        eval_every_n_epochs: int = 1,
        **kwargs,
    ):
        super(CalibModelCalibrator, self).__init__(*args, **kwargs)
        self._calib_outputs = calib_outputs
        self._calib_targets = calib_targets
        self._val_outputs = val_outputs
        self._val_targets = val_targets
        self.predict_fn = predict_fn
        self.uncertainty_fn = uncertainty_fn
        self.save_checkpoint_dir = save_checkpoint_dir
        self.save_every_n_steps = save_every_n_steps
        self.keep_top_n_checkpoints = keep_top_n_checkpoints
        self.disable_training_metrics_computation = disable_training_metrics_computation
        self.eval_every_n_epochs = eval_every_n_epochs
        self.multi_device = False

    def train(
        self,
        rng: PRNGKeyArray,
        state: CalibState,
        fun: Callable,
        n_epochs: int = 1,
        metrics: Optional[
            Tuple[Callable[[jnp.ndarray, jnp.ndarray, Array], Array], ...]
        ] = None,
        verbose: bool = True,
    ) -> Tuple[CalibState, Status]:
        training_losses_and_metrics = collections.defaultdict(list)
        val_losses_and_metrics = collections.defaultdict(list)

        state, targets, outputs, rng = self.on_train_start(
            state,
            [self._calib_targets, self._val_targets],
            [self._calib_outputs, self._val_outputs],
            rng,
        )
        calib_targets, val_targets = targets
        calib_outputs, val_outputs = outputs

        progress_bar = trange(n_epochs, desc="Epoch")
        for epoch in progress_bar:
            # training loop
            (
                state,
                training_losses_and_metrics_current_epoch,
                training_batch_metrics_str,
            ) = self._training_loop(
                epoch,
                fun,
                metrics,
                rng,
                state,
                calib_targets,
                calib_outputs,
                verbose,
                progress_bar,
            )
            # keep track of training losses and metrics [granularity=epoch]
            for k in training_losses_and_metrics_current_epoch.keys():
                training_losses_and_metrics[k].append(
                    training_losses_and_metrics_current_epoch[k]
                )

            # validation loop
            if self.should_perform_validation(val_targets, epoch):
                # performance evaluation on the whole validation dataset
                state = self.on_val_start(state)
                (
                    val_losses_and_metrics_current_epoch,
                    val_epoch_metrics_str,
                ) = self._val_loop(
                    fun=fun,
                    metrics=metrics,
                    rng=rng,
                    state=state,
                    targets=val_targets,
                    outputs=val_outputs,
                    verbose=verbose,
                )
                if verbose:
                    logging.info(f"Epoch: {epoch + 1} | " + val_epoch_metrics_str)
                # keep track of training losses and metrics [granularity=epoch] and check for early stopping
                for k in val_losses_and_metrics_current_epoch.keys():
                    val_losses_and_metrics[k].append(
                        val_losses_and_metrics_current_epoch[k]
                    )
                # check for early stopping
                if self.is_early_stopping_active and self._early_stopping.should_stop:
                    logging.info("[Early Stopping] Stopping training...")
                    break

        # aggregate
        training_status = {
            k: jnp.array(v) for k, v in training_losses_and_metrics.items()
        }
        val_status = {k: jnp.array(v) for k, v in val_losses_and_metrics.items()}
        status = dict(**training_status, **val_status)

        state = self.on_train_end(state)
        return state, status

    def _training_loop(
        self,
        current_epoch: int,
        fun: Callable,
        metrics: Optional[
            Tuple[Callable[[jnp.ndarray, jnp.ndarray, Array], Array], ...]
        ],
        rng: PRNGKeyArray,
        state: CalibState,
        targets: Array,
        outputs: Array,
        verbose: bool,
        progress_bar: TqdmDecorator,
    ) -> Tuple[CalibState, Dict[str, float], str]:
        training_losses_and_metrics_epoch_all_steps = []
        training_batch_metrics_str = ""
        # forward and backward pass
        state, aux = self.training_step(state, targets, outputs, fun, rng)
        # compute training losses and metrics for the current batch
        training_losses_and_metrics_current_batch = self.training_step_end(
            current_epoch=current_epoch,
            state=state,
            aux=aux,
            targets=targets,
            metrics=metrics,
        )
        # keep track of training losses and metrics [granularity=batch]
        training_losses_and_metrics_epoch_all_steps.append(
            training_losses_and_metrics_current_batch
        )
        # logging
        if verbose:
            training_batch_metrics_str = " | ".join(
                [
                    f"{m}: {round(float(v), 5)}"
                    for m, v in training_losses_and_metrics_current_batch.items()
                ]
            )
            progress_bar.set_description(
                f"Epoch: {current_epoch + 1} | " + training_batch_metrics_str,
                refresh=True,
            )

        # compute training losses and metrics avg for the current epoch + other ops (if needed)
        training_losses_and_metrics_current_epoch = self.training_epoch_end(
            training_losses_and_metrics_epoch_all_steps
        )

        return (
            state,
            training_losses_and_metrics_current_epoch,
            training_batch_metrics_str,
        )

    def training_step(
        self,
        state: CalibState,
        targets: Array,
        outputs: Array,
        fun: Callable,
        rng: PRNGKeyArray,
    ) -> Tuple[CalibState, Dict[str, Any]]:
        # ensure to use a different key at each step
        model_key = random.fold_in(rng, state.step)

        grad_fn = value_and_grad(
            lambda params: self.training_loss_step(
                fun, params, targets, outputs, state.mutable, model_key,
            ),
            has_aux=True,
        )
        (loss, aux), grad = grad_fn(state.params)
        grad, loss = self.sync_gradients_and_loss(grad, loss)

        state = state.apply_gradients(grads=grad, mutable=aux["mutable"])
        return (
            state,
            {
                "loss": loss,
                "outputs": aux["outputs"],
                "logging_kwargs": aux["logging_kwargs"],
            },
        )

    def training_loss_step(
        self,
        fun: Callable[[Any], Union[float, Tuple[float, dict]]],
        params: CalibParams,
        targets: Array,
        outputs: Array,
        mutable: CalibMutable,
        rng: PRNGKeyArray,
    ) -> Tuple[jnp.ndarray, Dict[str, Any]]:
        return_aux = ["outputs"]
        if mutable is not None:
            return_aux += ["mutable"]
        loss, aux = fun(
            params=params,
            targets=targets,
            outputs=outputs,
            mutable=mutable,
            rng=rng,
            return_aux=["outputs", "mutable"],
        )
        loss = -loss
        logging_kwargs = None
        return (
            loss,
            {
                "outputs": aux.get("outputs"),
                "mutable": aux.get("mutable"),
                "logging_kwargs": logging_kwargs,
            },
        )

    def training_step_end(
        self,
        current_epoch: int,
        state: CalibState,
        aux: Dict[str, Any],
        targets: Array,
        metrics: Optional[
            Tuple[Callable[[jnp.ndarray, jnp.ndarray, Array], Array], ...]
        ],
    ) -> Dict[str, jnp.ndarray]:
        if (
            self.save_checkpoint_dir
            and self.save_every_n_steps
            and current_epoch % self.save_every_n_steps == 0
        ):
            self.save_checkpoint(
                state, self.save_checkpoint_dir, keep=self.keep_top_n_checkpoints
            )
        training_losses_and_metrics = {"loss": aux["loss"]}

        if aux["logging_kwargs"] is not None:
            for k, v in aux["logging_kwargs"].items():
                training_losses_and_metrics[k] = v

        if not self.disable_training_metrics_computation and metrics is not None:
            preds = self.predict_fn(aux["outputs"])
            uncertainties = self.uncertainty_fn(aux["outputs"])
            if self.multi_device:
                training_batch_metrics = self.compute_metrics(
                    preds.reshape((preds.shape[0] * preds.shape[1],) + preds.shape[2:]),
                    uncertainties.reshape(
                        (uncertainties.shape[0] * uncertainties.shape[1],)
                        + uncertainties.shape[2:]
                    ),
                    targets.reshape(
                        (targets.shape[0] * targets.shape[1],) + targets.shape[2:]
                    ),
                    metrics,
                )
            else:
                training_batch_metrics = self.compute_metrics(
                    preds, uncertainties, targets, metrics
                )
            for k, v in training_batch_metrics.items():
                training_losses_and_metrics[k] = v
        return training_losses_and_metrics

    def _val_loop(
        self,
        fun: Callable,
        metrics: Optional[
            Tuple[Callable[[jnp.ndarray, jnp.ndarray, Array], Array], ...]
        ],
        rng: PRNGKeyArray,
        state: CalibState,
        targets: Array,
        outputs: Array,
        verbose: bool = True,
    ) -> Tuple[Dict[str, float], str]:
        val_losses_and_metrics_epoch_all_steps = []
        val_epoch_metrics_str = ""
        val_losses_and_metrics_current_batch = self.val_step(
            state, targets, outputs, fun, rng, metrics,
        )
        val_losses_and_metrics_epoch_all_steps.append(
            val_losses_and_metrics_current_batch
        )
        # compute validation losses and metrics for the current epoch
        val_losses_and_metrics_current_epoch = self.val_epoch_end(
            val_losses_and_metrics_epoch_all_steps, state
        )
        # logging
        if verbose:
            val_epoch_metrics_str = " | ".join(
                [
                    f"{m}: {round(float(v), 5)}"
                    for m, v in val_losses_and_metrics_current_epoch.items()
                ]
            )
        return val_losses_and_metrics_current_epoch, val_epoch_metrics_str

    def val_step(
        self,
        state: CalibState,
        targets: Array,
        outputs: Array,
        fun: Callable,
        rng: PRNGKeyArray,
        metrics: Optional[
            Tuple[Callable[[jnp.ndarray, jnp.ndarray, Array], Array], ...]
        ] = None,
    ) -> Dict[str, jnp.ndarray]:
        val_loss, aux = self.val_loss_step(state, targets, outputs, fun, rng)
        val_metrics = self.val_metrics_step(aux, targets, metrics)
        return {"val_loss": val_loss, **val_metrics}

    def val_loss_step(
        self,
        state: CalibState,
        targets: Array,
        outputs: Array,
        fun: Callable,
        rng: PRNGKeyArray,
    ) -> Tuple[jnp.ndarray, Dict[str, jnp.ndarray]]:
        log_joint_probs, aux = fun(
            params=state.params,
            targets=targets,
            outputs=outputs,
            mutable=state.mutable,
            rng=rng,
            return_aux=["outputs"],
        )
        return -log_joint_probs, aux

    def val_metrics_step(
        self,
        aux: Dict[str, jnp.ndarray],
        targets: Array,
        metrics: Optional[
            Tuple[Callable[[jnp.ndarray, jnp.ndarray, Array], Array], ...]
        ] = None,
    ) -> Dict[str, jnp.ndarray]:
        if metrics is not None:
            val_metrics = self.compute_metrics(
                self.predict_fn(aux["outputs"]),
                self.uncertainty_fn(aux["outputs"]),
                targets,
                metrics,
            )
            return {f"val_{m}": v for m, v in val_metrics.items()}
        else:
            return {}

    def training_epoch_end(
        self, training_losses_and_metrics_current_epoch: List[Dict[str, jnp.ndarray]]
    ) -> Dict[str, float]:
        return self._get_mean_losses_and_metrics(
            training_losses_and_metrics_current_epoch
        )

    def val_epoch_end(
        self,
        val_losses_and_metrics_current_epoch: List[Dict[str, jnp.ndarray]],
        state: CalibState,
    ) -> Dict[str, float]:
        val_losses_and_metrics_current_epoch = self._get_mean_losses_and_metrics(
            val_losses_and_metrics_current_epoch
        )
        # early stopping
        improved = self.early_stopping_update(val_losses_and_metrics_current_epoch)
        if improved and self.save_checkpoint_dir:
            self.save_checkpoint(state, self.save_checkpoint_dir, force_save=True)
        return val_losses_and_metrics_current_epoch

    def _get_mean_losses_and_metrics(
        self, losses_and_metrics: List[Dict[str, jnp.ndarray]]
    ) -> Dict[str, float]:
        losses_and_metrics = stack_forest(losses_and_metrics)
        losses_and_metrics = tree_map(lambda x: x.mean(), losses_and_metrics)
        return losses_and_metrics

    def should_perform_validation(
        self, val_data_loader: Optional[DataLoader], epoch: int
    ) -> bool:
        return (
            val_data_loader is not None
            and self.eval_every_n_epochs > 0
            and epoch % self.eval_every_n_epochs == 0
        )

    @staticmethod
    def sync_gradients_and_loss(
        grad: jnp.ndarray, loss: jnp.ndarray
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        return grad, loss

    def on_train_start(
        self,
        state: CalibState,
        targets: List[Array],
        outputs: List[Array],
        rng: PRNGKeyArray,
    ) -> Tuple[CalibState, List[Array], List[Array], PRNGKeyArray]:
        return state, targets, outputs, rng

    def on_train_end(self, state: CalibState) -> CalibState:
        self.save_checkpoint(
            state,
            save_checkpoint_dir=self.save_checkpoint_dir,
            keep=self.keep_top_n_checkpoints,
            force_save=True,
        )
        return state

    def on_val_start(self, state: CalibState) -> CalibState:
        return state

    def compute_metrics(
        self,
        preds: Array,
        uncertainties: Array,
        targets: Array,
        metrics: Optional[
            Tuple[Callable[[jnp.ndarray, jnp.ndarray, Array], Array], ...]
        ],
    ) -> Dict[str, Array]:
        metrics_vals = {}
        for metric in metrics:
            metrics_vals[metric.__name__] = metric(preds, uncertainties, targets)
        return metrics_vals


class JittedMixin:
    @partial(jax.jit, static_argnums=(0, 4))
    def training_step(
        self,
        state: CalibState,
        targets: Array,
        outputs: Array,
        fun: Callable,
        rng: PRNGKeyArray,
    ) -> Tuple[CalibState, Dict[str, Any]]:
        return super().training_step(state, targets, outputs, fun, rng)

    @partial(jax.jit, static_argnums=(0, 4))
    def val_loss_step(
        self,
        state: CalibState,
        targets: Array,
        outputs: Array,
        fun: Callable,
        rng: PRNGKeyArray,
    ) -> Dict[str, jnp.ndarray]:
        return super().val_loss_step(state, targets, outputs, fun, rng)


class MultiDeviceMixin:
    all_reduce_mean = jax.pmap(lambda x: lax.pmean(x, "x"), "x")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.multi_device = True

    @staticmethod
    def _add_device_dim_to_array(arr: Array) -> Array:
        n_devices = jax.local_device_count()
        if arr.shape[0] % n_devices != 0:
            raise ValueError(
                f"The number of data points of all outputs and targets must be a multiple of {n_devices}, "
                f"that is the number of available devices. However, {arr.shape[0]} were found."
            )
        return arr.reshape((n_devices, -1) + arr.shape[1:]) if arr is not None else arr

    @staticmethod
    def sync_mutable(state: CalibState) -> CalibState:
        return (
            state.replace(mutable=MultiDeviceMixin.all_reduce_mean(state.mutable))
            if state.mutable["output_calibrator"] is not None
            else state
        )

    @staticmethod
    def sync_gradients_and_loss(
        grads: jnp.ndarray, loss: jnp.ndarray
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
        grad = lax.pmean(grads, axis_name="batch")
        loss = lax.pmean(loss, axis_name="batch")
        return grad, loss

    def save_checkpoint(
        self,
        state: CalibState,
        save_checkpoint_dir: Path,
        keep: int = 1,
        force_save: bool = False,
        prefix: str = "checkpoint_",
    ) -> None:
        state = self.sync_mutable(state)
        state = jax.device_get(tree_map(lambda x: x[0], state))
        return super(MultiDeviceMixin, self).save_checkpoint(
            state, save_checkpoint_dir, keep, force_save, prefix
        )

    def on_train_start(
        self,
        state: CalibState,
        targets: List[Array],
        outputs: List[Array],
        rng: PRNGKeyArray,
    ) -> Tuple[CalibState, List[DataLoader], List[TargetsLoader], PRNGKeyArray]:
        state, targets, outputs, rng = super(MultiDeviceMixin, self).on_train_start(
            state, targets, outputs, rng
        )
        state = jax_utils.replicate(state)
        targets = [
            self._add_device_dim_to_array(arr) if arr is not None else arr
            for arr in targets
        ]
        outputs = [
            self._add_device_dim_to_array(arr) if arr is not None else arr
            for arr in outputs
        ]
        model_key = random.split(rng, jax.local_device_count())
        return state, targets, outputs, model_key

    def on_train_end(self, state: CalibState) -> CalibState:
        state = super(MultiDeviceMixin, self).on_train_end(state)
        return jax.device_get(tree_map(lambda x: x[0], state))

    @partial(jax.pmap, axis_name="batch", static_broadcasted_argnums=(0, 4))
    def training_step(
        self,
        state: CalibState,
        targets: Array,
        outputs: Array,
        fun: Callable,
        rng: PRNGKeyArray,
    ) -> Tuple[CalibState, Dict[str, Any]]:
        return super().training_step(state, targets, outputs, fun, rng)

    def training_step_end(
        self,
        current_epoch: int,
        state: CalibState,
        aux: Dict[str, Any],
        targets: Array,
        metrics: Optional[Tuple[Callable[[jnp.ndarray, Array], Array], ...]],
    ) -> Dict[str, jnp.ndarray]:
        training_losses_and_metrics = super(MultiDeviceMixin, self).training_step_end(
            current_epoch, state, aux, targets, metrics
        )
        return tree_map(lambda x: x.mean(), training_losses_and_metrics)

    def on_val_start(self, state: CalibState) -> CalibState:
        state = super(MultiDeviceMixin, self).on_val_start(state)
        if state.mutable["output_calibrator"] is not None:
            state = self.sync_mutable(state)
        return state

    @partial(jax.pmap, axis_name="batch", static_broadcasted_argnums=(0, 4))
    def val_loss_step(
        self,
        state: CalibState,
        targets: Array,
        outputs: Array,
        fun: Callable,
        rng: PRNGKeyArray,
    ) -> Dict[str, jnp.ndarray]:
        val_losses = super().val_loss_step(state, targets, outputs, fun, rng)
        return lax.pmean(val_losses, axis_name="batch")

    def val_metrics_step(
        self,
        aux: Dict[str, jnp.ndarray],
        targets: Array,
        metrics: Optional[
            Tuple[Callable[[jnp.ndarray, jnp.ndarray, Array], Array], ...]
        ] = None,
    ) -> Dict[str, jnp.ndarray]:
        outputs = aux["outputs"]
        outputs = outputs.reshape(outputs.shape[0] * outputs.shape[1], -1)
        targets = targets.reshape(targets.shape[0] * targets.shape[1], -1)
        if metrics is not None:
            val_metrics = self.compute_metrics(
                self.predict_fn(outputs), self.uncertainty_fn(outputs), targets, metrics
            )
            return {f"val_{m}": v for m, v in val_metrics.items()}
        else:
            return {}


class JittedCalibModelCalibrator(JittedMixin, CalibModelCalibrator):
    pass


class MultiDeviceCalibModelCalibrator(MultiDeviceMixin, CalibModelCalibrator):
    pass



class CalibProcessor:
    def __init__(
        self, devices: int = -1, disable_jit: bool = False,
    ):
        """
        An object to configure computational aspects of the calibration process.

        Parameters
        ----------
        devices: int
            A list of devices to be used during training.
            At the moment two options are supported: use all devices (`devices=-1`) or use no device (`devices=0`).
        disable_jit: bool
            if True, no function within the calibration loop is jitted.
        """
        self.devices = devices
        self.disable_jit = disable_jit



from __future__ import annotations

import logging
from typing import Optional

import jax.numpy as jnp
from jax import random
from jax._src.prng import PRNGKeyArray
from jax.flatten_util import ravel_pytree

from fortuna.data.loader import DataLoader, InputsLoader
from fortuna.prob_model.fit_config import FitConfig
from fortuna.prob_model.joint.base import Joint
from fortuna.prob_model.joint.state import JointState
from fortuna.prob_model.posterior.base import Posterior
from fortuna.prob_model.posterior.map.map_approximator import \
    MAPPosteriorApproximator
from fortuna.prob_model.posterior.map.map_posterior import MAPPosterior
from fortuna.prob_model.posterior.map.map_state import MAPState
from fortuna.prob_model.posterior.posterior_state_repository import \
    PosteriorStateRepository
from fortuna.prob_model.posterior.swag import SWAG_NAME
from fortuna.prob_model.posterior.swag.swag_approximator import \
    SWAGPosteriorApproximator
from fortuna.prob_model.posterior.swag.swag_state import SWAGState
from fortuna.prob_model.posterior.swag.swag_trainer import (
    JittedSWAGTrainer, MultiDeviceSWAGTrainer, SWAGTrainer)
from fortuna.typing import Array, Status
from fortuna.utils.device import select_trainer_given_devices


class SWAGPosterior(Posterior):
    def __init__(self, joint: Joint, posterior_approximator: SWAGPosteriorApproximator):
        """
        SWAG approximate posterior class.

        Parameters
        ----------
        joint: Joint
            A joint distribution object.
        posterior_approximator: SWAGPosteriorApproximator
            A SWAG posterior approximator.
        """
        super().__init__(joint=joint, posterior_approximator=posterior_approximator)

    def __str__(self):
        return SWAG_NAME

    def fit(
        self,
        train_data_loader: DataLoader,
        val_data_loader: Optional[DataLoader] = None,
        fit_config: FitConfig = FitConfig(),
        map_fit_config: Optional[FitConfig] = None,
        *kwargs,
    ) -> Status:
        if (
            fit_config.checkpointer.dump_state is True
            and not fit_config.checkpointer.save_checkpoint_dir
        ):
            raise ValueError(
                "`save_checkpoint_dir` must be passed when `dump_state` is set to True."
            )
        if self.posterior_approximator.rank < 2:
            raise ValueError("`rank` must be at least 2.")
        if fit_config.optimizer.n_epochs <= self.posterior_approximator.rank:
            raise ValueError(
                """Not enough SWAG epochs to obtain `rank={}`. Please either increase `n_swag_epochs` or 
            decrease `rank`.""".format(
                    self.posterior_approximator.rank
                )
            )
        if (
            fit_config.monitor.early_stopping_patience
            and fit_config.monitor.early_stopping_patience > 0
        ):
            logging.warning(
                f"""It seems you are trying to enable early stopping, since 
            `fit_config.monitor.early_stopping_patience={fit_config.monitor.early_stopping_patience}`. We do not 
            support early stopping in SWAG, since we implement it as a post-processing step of MAP. If your intention
            was rather to enable early stopping in MAP, please configure `map_fit_config` accordingly."""
            )

        init_prob_model_state, n_train_data, n_val_data = self._init(
            train_data_loader, val_data_loader
        )

        status = {}

        if not fit_config.checkpointer.restore_checkpoint_path:
            map_posterior = MAPPosterior(
                self.joint, posterior_approximator=MAPPosteriorApproximator()
            )
            map_posterior.rng = self.rng
            logging.info("Do a preliminary run of MAP.")
            status["map"] = map_posterior.fit(
                train_data_loader=train_data_loader,
                val_data_loader=val_data_loader,
                fit_config=map_fit_config
                if map_fit_config is not None
                else FitConfig(),
            )
            state = SWAGState.convert_from_map_state(
                map_state=map_posterior.state.get(),
                optimizer=fit_config.optimizer.method,
            )
            logging.info("Preliminary run with MAP completed.")
        else:
            state = self.restore_checkpoint(
                restore_checkpoint_path=fit_config.checkpointer.restore_checkpoint_path,
                optimizer=fit_config.optimizer.method,
            )
            if type(state) == MAPState:
                state = SWAGState.convert_from_map_state(
                    map_state=state, optimizer=fit_config.optimizer.method
                )

        trainer_cls = select_trainer_given_devices(
            devices=fit_config.processor.devices,
            BaseTrainer=SWAGTrainer,
            JittedTrainer=JittedSWAGTrainer,
            MultiDeviceTrainer=MultiDeviceSWAGTrainer,
            disable_jit=fit_config.processor.disable_jit,
        )
        trainer = trainer_cls(
            predict_fn=self.joint.likelihood.prob_output_layer.predict,
            save_checkpoint_dir=fit_config.checkpointer.save_checkpoint_dir,
            save_every_n_steps=fit_config.checkpointer.save_every_n_steps,
            keep_top_n_checkpoints=fit_config.checkpointer.keep_top_n_checkpoints,
            disable_training_metrics_computation=fit_config.monitor.disable_training_metrics_computation,
            eval_every_n_epochs=fit_config.monitor.eval_every_n_epochs,
            early_stopping_verbose=False,
        )

        kwargs = dict(rank=self.posterior_approximator.rank)
        logging.info("Run SWAG.")
        state, status["swag"] = trainer.train(
            rng=self.rng.get(),
            state=state,
            fun=self.joint._batched_log_joint_prob,
            training_dataloader=train_data_loader,
            training_dataset_size=n_train_data,
            n_epochs=fit_config.optimizer.n_epochs,
            metrics=fit_config.monitor.metrics,
            validation_dataloader=val_data_loader,
            validation_dataset_size=n_val_data,
            verbose=fit_config.monitor.verbose,
            **kwargs,
        )

        self.state = PosteriorStateRepository(
            fit_config.checkpointer.save_checkpoint_dir
            if fit_config.checkpointer.dump_state is True
            else None
        )
        self.state.put(state, keep=fit_config.checkpointer.keep_top_n_checkpoints)
        logging.info("Fit completed.")
        return status

    def sample(
        self,
        rng: Optional[PRNGKeyArray] = None,
        inputs_loader: Optional[InputsLoader] = None,
        inputs: Optional[Array] = None,
        **kwargs,
    ) -> JointState:
        """
        Sample from the posterior distribution.

        Parameters
        ----------
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        inputs_loader: Optional[InputsLoader]
            Input data loader. This or `inputs` is required if the posterior state includes mutable objects.
        inputs: Optional[Array]
            Input variables. This or `inputs_loader` is required if the posterior state includes mutable objects.

        Returns
        -------
        JointState
            A sample from the posterior distribution.
        """
        if rng is None:
            rng = self.rng.get()
        state = self.state.get()
        if state.mutable is not None and inputs_loader is None and inputs is None:
            raise ValueError(
                "The posterior state contains mutable objects. Please pass `inputs_loader` or `inputs`."
            )

        n_params = len(state.mean)
        unravel = ravel_pytree(state.params)[1]
        coeff1 = 1 / jnp.sqrt(2)
        coeff2 = coeff1 / jnp.sqrt(self.posterior_approximator.rank - 1)

        rng, key1, key2 = random.split(rng, 3)
        z1 = random.normal(key1, shape=(n_params,))
        z2 = random.normal(key2, shape=(self.posterior_approximator.rank,))
        state = state.replace(
            params=unravel(
                state.mean
                + coeff1 * state.std * z1
                + coeff2 * jnp.matmul(state.dev, z2)
            )
        )

        if state.mutable:
            if inputs_loader is not None:
                for batch_inputs in inputs_loader:
                    state = state.replace(
                        mutable=self.joint.likelihood.model_manager.apply(
                            state.params,
                            batch_inputs,
                            mutable=state.mutable,
                            train=True,
                            rng=rng,
                        )[1]["mutable"]
                    )
            else:
                state = state.replace(
                    mutable=self.joint.likelihood.model_manager.apply(
                        state.params, inputs, mutable=state.mutable, train=True, rng=rng
                    )[1]["mutable"]
                )

        return JointState(
            params=state.params,
            mutable=state.mutable,
            calib_params=state.calib_params,
            calib_mutable=state.calib_mutable,
        )



from __future__ import annotations

from typing import List, Tuple, Union

import jax.numpy as jnp
from flax.core import FrozenDict

from fortuna.prob_model.posterior.map.map_state import MAPState
from fortuna.prob_model.posterior.state import PosteriorState
from fortuna.typing import Params
from fortuna.utils.nested_dicts import nested_pair
from fortuna.utils.strings import convert_string_to_jnp_array


class LaplaceState(PosteriorState):
    """
    Attributes
    ----------
    encoded_name: jnp.ndarray
        Laplace state name encoded as an array.
    """

    encoded_name: jnp.ndarray = convert_string_to_jnp_array("LaplaceState")

    @classmethod
    def convert_from_map_state(
        cls,
        map_state: MAPState,
        std: Union[Params, Tuple[Params, ...]],
        which_params: Tuple[List],
    ) -> LaplaceState:
        """
        Convert a MAP state into a Laplace state.

        Parameters
        ----------
        map_state: MAPState
            A MAP state.
        std: Union[Params, Tuple[Params, ...]]
            Standard deviation parameters.
        which_params: Tuple[List]
            Sequences of keys pointing to the parameters over which `std` is defined. If `which_params` is None,
            `std` must be defined for all parameters.

        Returns
        -------
        LaplaceState
            A Laplace state instance.
        """
        params = map_state.params.unfreeze()
        if which_params:
            params = FrozenDict(
                nested_pair(params, which_params, std, ("mean", "std"),)
            )
        else:
            for k, v in params.items():
                params[k] = FrozenDict(
                    {"params": dict(mean=v["params"], std=std[k]["params"])}
                )
        map_state = map_state.replace(
            params=FrozenDict(params), encoded_name=LaplaceState.encoded_name
        )
        return LaplaceState.init_from_dict(vars(map_state))



from fortuna.prob_model.posterior.normalizing_flow.advi import ADVI_NAME
from fortuna.prob_model.posterior.normalizing_flow.normalizing_flow_trainer import \
    NormalizingFlowTrainer


class ADVITrainer(NormalizingFlowTrainer):
    def __str__(self):
        return ADVI_NAME



from fortuna.calibration.calibrator import CalibTrainerABC
from fortuna.calibration.calib_state import CalibState
from typing import Callable, Any, Union, Tuple, Optional, Dict
import jax.numpy as jnp
from flax.core import FrozenDict
from fortuna.typing import CalibMutable, CalibParams, Batch, Array
from optax._src.base import PyTree
from jax._src.prng import PRNGKeyArray


class ProbCalibTrainer(CalibTrainerABC):
    def training_loss_step(
        self,
        fun: Callable[[Any], Union[float, Tuple[float, dict]]],
        params: CalibParams,
        batch: Batch,
        mutable: CalibMutable,
        rng: PRNGKeyArray,
        n_data: int,
    ) -> Tuple[jnp.ndarray, Dict[str, Any]]:
        return_aux = ["outputs"]
        if mutable is not None:
            return_aux += ["mutable"]
        loss, aux = fun(
            batch,
            n_data=n_data,
            return_aux=["outputs", "calib_mutable"],
            ensemble_outputs=self.outputs,
            calib_params=params,
            calib_mutable=mutable,
            rng=rng,
        )
        loss = -loss
        logging_kwargs = None
        return (
            loss,
            {
                "outputs": aux.get("outputs"),
                "mutable": aux.get("mutable"),
                "logging_kwargs": logging_kwargs,
            },
        )

    def validation_step(
        self,
        state: CalibState,
        batch: Batch,
        fun: Callable,
        rng: PRNGKeyArray,
        n_data: int,
        metrics: Optional[Tuple[Callable[[jnp.ndarray, jnp.ndarray, Array], float], ...]] = None,
        unravel: Optional[Callable[[any], PyTree]] = None,
        kwargs: FrozenDict[str, Any] = FrozenDict(),
    ) -> Dict[str, jnp.ndarray]:
        log_probs, aux = fun(
            batch,
            n_data=n_data,
            return_aux=["outputs"],
            ensemble_outputs=self.outputs,
            calib_params=state.params,
            calib_mutable=state.mutable,
            rng=rng,
        )

        if metrics is not None:
            val_metrics = self.compute_metrics(
                self.predict_fn(aux["outputs"]), self.uncertainty_fn(aux["outputs"]), batch[1], metrics
            )
            return {
                "val_loss": -log_probs,
                **{f"val_{m}": v for m, v in val_metrics.items()},
            }
        return dict(val_loss=-log_probs)

    def __str__(self):
        return "calibration"


class JittedProbCalibTrainer(JittedMixin, ProbCalibTrainer):
    pass


class MultiGPUMAPTrainer(MultiGPUMixin, ProbCalibTrainer):
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fortuna',
 'fortuna.calib_config',
 'fortuna.calib_model',
 'fortuna.calib_model.calib_config',
 'fortuna.calib_model.predictive',
 'fortuna.calibration',
 'fortuna.conformal',
 'fortuna.data',
 'fortuna.distribution',
 'fortuna.metric',
 'fortuna.model',
 'fortuna.model.model_manager',
 'fortuna.output_calibrator',
 'fortuna.output_calibrator.output_calib_manager',
 'fortuna.prob_model',
 'fortuna.prob_model.calib_config',
 'fortuna.prob_model.fit_config',
 'fortuna.prob_model.joint',
 'fortuna.prob_model.likelihood',
 'fortuna.prob_model.posterior',
 'fortuna.prob_model.posterior.deep_ensemble',
 'fortuna.prob_model.posterior.laplace',
 'fortuna.prob_model.posterior.map',
 'fortuna.prob_model.posterior.normalizing_flow',
 'fortuna.prob_model.posterior.normalizing_flow.advi',
 'fortuna.prob_model.posterior.swag',
 'fortuna.prob_model.predictive',
 'fortuna.prob_model.prior',
 'fortuna.prob_output_layer',
 'fortuna.training',
 'fortuna.utils']

package_data = \
{'': ['*']}

install_requires = \
['flax>=0.6.2,<0.7.0',
 'matplotlib>=3.6.2,<4.0.0',
 'numpy>=1.23.4,<2.0.0',
 'optax>=0.1.3,<0.2.0',
 'tensorflow-cpu>=2.11.0,<3.0.0',
 'tqdm>=4.64.1,<5.0.0']

extras_require = \
{'docs': ['Sphinx>=5.3.0,<6.0.0',
          'sphinx-autodoc-typehints>=1.19.5,<2.0.0',
          'nbsphinx>=0.8.10,<0.9.0',
          'nbsphinx-link>=1.3.0,<2.0.0',
          'sphinx-gallery>=0.11.1,<0.12.0',
          'pydata-sphinx-theme>=0.12.0,<0.13.0',
          'ipython>=8.7.0,<9.0.0']}

setup_kwargs = {
    'name': 'aws-fortuna',
    'version': '0.1.5',
    'description': 'A Library for Uncertainty Quantification.',
    'long_description': 'Fortuna\n#######\n\n.. image:: https://img.shields.io/pypi/status/Fortuna\n    :target: https://img.shields.io/pypi/status/Fortuna\n    :alt: PyPI - Status\n.. image:: https://img.shields.io/pypi/dm/aws-fortuna\n    :target: https://pypistats.org/packages/aws-fortuna\n    :alt: PyPI - Downloads\n.. image:: https://img.shields.io/pypi/v/aws-fortuna\n    :target: https://img.shields.io/pypi/v/aws-fortuna\n    :alt: PyPI - Version\n.. image:: https://img.shields.io/github/license/awslabs/Fortuna\n    :target: https://github.com/awslabs/Fortuna/blob/main/LICENSE\n    :alt: License\n.. image:: https://readthedocs.org/projects/aws-fortuna/badge/?version=latest\n    :target: https://aws-fortuna.readthedocs.io\n    :alt: Documentation Status\n\nA Library for Uncertainty Quantification\n========================================\nProper estimation of predictive uncertainty is fundamental in applications that involve critical decisions.\nUncertainty can be used to assess reliability of model predictions, trigger human intervention,\nor decide whether a model can be safely deployed in the wild.\n\nFortuna is a library for uncertainty quantification that makes it easy for users to run benchmarks and bring uncertainty to production systems.\nFortuna provides calibration and conformal methods starting from pre-trained models written in any framework,\nand it further supports several Bayesian inference methods starting from deep learning models written in `Flax <https://flax.readthedocs.io/en/latest/index.html>`_.\nThe language is designed to be intuitive for practitioners unfamiliar with uncertainty quantification,\nand is highly configurable.\n\nCheck the `documentation <https://aws-fortuna.readthedocs.io/en/latest/>`_ for a quickstart, examples and references.\n\nUsage modes\n===========\nFortuna offers three different usage modes:\n`From uncertainty estimates <https://github.com/awslabs/fortuna#from-uncertainty-estimates>`_,\n`From model outputs <https://github.com/awslabs/fortuna#from-model-outputs>`_ and\n`From Flax models <https://github.com/awslabs/fortuna#from-flax-models>`_.\nThese serve users according to the constraints dictated by their own applications.\nTheir pipelines are depicted in the following figure, each starting from one of the green panels.\n\n.. image:: https://github.com/awslabs/fortuna/raw/main/docs/source/_static/pipeline.png\n    :target: https://github.com/awslabs/fortuna/raw/main/docs/source/_static/pipeline.png\n\nFrom uncertainty estimates\n---------------------------\nStarting from uncertainty estimates has minimal compatibility requirements and it is the quickest level of interaction with the library.\nThis usage mode offers conformal prediction methods for both classification and regression.\nThese take uncertainty estimates in input,\nand return rigorous sets of predictions that retain a user-given level of probability.\nIn one-dimensional regression tasks, conformal sets may be thought as calibrated versions of confidence or credible intervals.\n\nMind that if the uncertainty estimates that you provide in inputs are inaccurate,\nconformal sets might be large and unusable.\nFor this reason, if your application allows it,\nplease consider the `From model outputs <https://github.com/awslabs/fortuna#from-model-outputs>`_ and\n`From Flax models <https://github.com/awslabs/fortuna#from-flax-models>`_ usage modes.\n\n**Example.** Suppose you want to calibrate credible intervals with coverage error :code:`error`,\neach corresponding to a different test input variable.\nWe assume that credible intervals are passed as arrays of lower and upper bounds,\nrespectively :code:`test_lower_bounds` and :code:`test_upper_bounds`.\nYou also have lower and upper bounds of credible intervals computed for several validation inputs,\nrespectively :code:`val_lower_bounds` and :code:`val_upper_bounds`.\nThe corresponding array of validation targets is denoted by :code:`val_targets`.\nThe following code produces *conformal prediction intervals*,\ni.e. calibrated versions of you test credible intervals.\n\n.. code-block:: python\n\n from fortuna.conformal.regression import QuantileConformalRegressor\n conformal_intervals = QuantileConformalRegressor().conformal_interval(\n      val_lower_bounds=val_lower_bounds, val_upper_bounds=val_upper_bounds,\n      test_lower_bounds=test_lower_bounds, test_upper_bounds=test_upper_bounds,\n      val_targets=val_targets, error=error)\n\nFrom model outputs\n------------------\nStarting from model outputs assumes you have already trained a model in some framework,\nand arrive to Fortuna with model outputs in :code:`numpy.ndarray` format for each input data point.\nThis usage mode allows you to calibrate your model outputs, estimate uncertainty,\ncompute metrics and obtain conformal sets.\n\nCompared to the `From uncertainty estimates <https://github.com/awslabs/fortuna#from-uncertainty-estimates>`_ usage mode,\nthis one offers better control,\nas it can make sure uncertainty estimates have been appropriately calibrated.\nHowever, if the model had been trained with classical methods,\nthe resulting quantification of model (a.k.a. epistemic) uncertainty may be poor.\nTo mitigate this problem, please consider the `From Flax models <https://github.com/awslabs/fortuna#from-flax-models>`_\nusage mode.\n\n**Example.**\nSuppose you have validation and test model outputs,\nrespectively :code:`val_outputs` and :code:`test_outputs`.\nFurthermore, you have some arrays of validation and target variables,\nrespectively :code:`val_targets` and :code:`test_targets`.\nThe following code provides a minimal classification example to get calibrated predictive entropy estimates.\n\n.. code-block:: python\n\n  from fortuna.calib_model import CalibClassifier\n  calib_model = CalibClassifier()\n  status = calib_model.calibrate(outputs=val_outputs, targets=val_targets)\n  test_entropies = calib_model.predictive.entropy(outputs=test_outputs)\n\nFrom Flax models\n--------------------------\nStarting from Flax models has higher compatibility requirements than the\n`From uncertainty estimates <https://github.com/awslabs/fortuna#from-uncertainty-estimates>`_\nand `From model outputs <https://github.com/awslabs/fortuna#from-model-outputs>`_ usage modes,\nas it requires deep learning models written in `Flax <https://flax.readthedocs.io/en/latest/index.html>`_.\nHowever, it enables you to replace standard model training with scalable Bayesian inference procedures,\nwhich may significantly improve the quantification of predictive uncertainty.\n\n**Example.** Suppose you have a Flax classification deep learning model :code:`model` from inputs to logits, with output\ndimension given by :code:`output_dim`. Furthermore,\nyou have some training, validation and calibration TensorFlow data loader :code:`train_data_loader`, :code:`val_data_loader`\nand :code:`test_data_loader`, respectively.\nThe following code provides a minimal classification example to get calibrated probability estimates.\n\n.. code-block:: python\n\n  from fortuna.data import DataLoader\n  train_data_loader = DataLoader.from_tensorflow_data_loader(train_data_loader)\n  calib_data_loader = DataLoader.from_tensorflow_data_loader(val_data_loader)\n  test_data_loader = DataLoader.from_tensorflow_data_loader(test_data_loader)\n\n  from fortuna.prob_model import ProbClassifier\n  prob_model = ProbClassifier(model=model)\n  status = prob_model.train(train_data_loader=train_data_loader, calib_data_loader=calib_data_loader)\n  test_means = prob_model.predictive.mean(inputs_loader=test_data_loader.to_inputs_loader())\n\n\nInstallation\n============\n**NOTE:** Before installing Fortuna, you are required to `install JAX <https://github.com/google/jax#installation>`_ in your virtual environment.\n\nYou can install Fortuna by typing\n\n.. code-block::\n\n    pip install aws-fortuna\n\nAlternatively, you can build the package using `Poetry <https://python-poetry.org/docs/>`_.\nIf you choose to pursue this way, first install Poetry and add it to your PATH\n(see `here <https://python-poetry.org/docs/#installation>`_). Then type\n\n.. code-block::\n\n    poetry install\n\nAll the dependecies will be installed at their required versions.\nIf you also want to install the optional Sphinx dependencies to build the documentation,\nadd the flag :code:`-E docs` to the command above.\nFinally, you can either access the virtualenv that Poetry created by typing :code:`poetry shell`,\nor execute commands within the virtualenv using the :code:`run` command, e.g. :code:`poetry run python`.\n\nExamples\n========\nSeveral usage examples are found in the\n`/examples <https://github.com/awslabs/fortuna/tree/main/examples>`_\ndirectory.\n\nContributing\n============\nIf you wish to contribute to the project, please refer to our `contribution guidelines <https://github.com/awslabs/fortuna/blob/main/CONTRIBUTING.md>`_.\n\n\nLicense\n=======\nThis project is licensed under the Apache-2.0 License.\nSee `LICENSE <https://github.com/awslabs/fortuna/blob/main/LICENSE>`_ for more information.\n',
    'author': 'Gianluca Detommaso',
    'author_email': 'gianluca.detommaso@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from __future__ import annotations

import logging
import os
import pathlib
from typing import List, Optional

import numpy as np
from jax._src.prng import PRNGKeyArray

from fortuna.data.loader import DataLoader
from fortuna.prob_model.fit_config import FitConfig
from fortuna.prob_model.joint.base import Joint
from fortuna.prob_model.joint.state import JointState
from fortuna.prob_model.posterior.base import Posterior
from fortuna.prob_model.posterior.deep_ensemble import DEEP_ENSEMBLE_NAME
from fortuna.prob_model.posterior.deep_ensemble.deep_ensemble_approximator import \
    DeepEnsemblePosteriorApproximator
from fortuna.prob_model.posterior.deep_ensemble.deep_ensemble_repositories import \
    DeepEnsemblePosteriorStateRepository
from fortuna.prob_model.posterior.map.map_posterior import MAPState
from fortuna.prob_model.posterior.map.map_trainer import (
    JittedMAPTrainer, MAPTrainer, MultiDeviceMAPTrainer)
from fortuna.typing import Path, Status
from fortuna.utils.device import select_trainer_given_devices

logger = logging.getLogger(__name__)


class DeepEnsemblePosterior(Posterior):
    def __init__(
        self, joint: Joint, posterior_approximator: DeepEnsemblePosteriorApproximator,
    ):
        """
        Deep ensemble approximate posterior class.

        Parameters
        ----------
        joint: Joint
            Joint distribution.
        posterior_approximator: DeepEnsemble
            Deep ensemble posterior approximator.
        """
        super().__init__(joint=joint, posterior_approximator=posterior_approximator)

    def __str__(self):
        return DEEP_ENSEMBLE_NAME

    def fit(
        self,
        train_data_loader: DataLoader,
        val_data_loader: Optional[DataLoader] = None,
        fit_config: FitConfig = FitConfig(),
        **kwargs,
    ) -> List[Status]:
        if (
            fit_config.checkpointer.dump_state is True
            and not fit_config.checkpointer.save_checkpoint_dir
        ):
            raise ValueError(
                "`save_checkpoint_dir` must be passed when `dump_state` is set to True."
            )

        trainer_cls = select_trainer_given_devices(
            devices=fit_config.processor.devices,
            BaseTrainer=MAPTrainer,
            JittedTrainer=JittedMAPTrainer,
            MultiDeviceTrainer=MultiDeviceMAPTrainer,
            disable_jit=fit_config.processor.disable_jit,
        )

        def _fit(i):
            init_prob_model_state, n_train_data, n_val_data = self._init(
                train_data_loader, val_data_loader
            )

            save_checkpoint_dir_i = (
                pathlib.Path(fit_config.checkpointer.save_checkpoint_dir) / str(i)
                if fit_config.checkpointer.save_checkpoint_dir
                else None
            )
            trainer = trainer_cls(
                predict_fn=self.joint.likelihood.prob_output_layer.predict,
                save_checkpoint_dir=save_checkpoint_dir_i,
                save_every_n_steps=fit_config.checkpointer.save_every_n_steps,
                keep_top_n_checkpoints=fit_config.checkpointer.keep_top_n_checkpoints,
                disable_training_metrics_computation=fit_config.monitor.disable_training_metrics_computation,
                eval_every_n_epochs=fit_config.monitor.eval_every_n_epochs,
                early_stopping_monitor=fit_config.monitor.early_stopping_monitor,
                early_stopping_min_delta=fit_config.monitor.early_stopping_min_delta,
                early_stopping_patience=fit_config.monitor.early_stopping_patience,
            )

            if fit_config.checkpointer.restore_checkpoint_path:
                state = self.restore_checkpoint(
                    restore_checkpoint_path=str(
                        fit_config.checkpointer.restore_checkpoint_path
                    )
                    + "/"
                    + str(i),
                    optimizer=fit_config.optimizer.method,
                )
            else:
                state = MAPState.init(
                    params=init_prob_model_state.params,
                    mutable=init_prob_model_state.mutable,
                    optimizer=fit_config.optimizer.method,
                    calib_params=init_prob_model_state.calib_params,
                    calib_mutable=init_prob_model_state.calib_mutable,
                )

            return trainer.train(
                rng=self.rng.get(),
                state=state,
                fun=self.joint._batched_log_joint_prob,
                training_dataloader=train_data_loader,
                training_dataset_size=n_train_data,
                n_epochs=fit_config.optimizer.n_epochs,
                metrics=fit_config.monitor.metrics,
                validation_dataloader=val_data_loader,
                validation_dataset_size=n_val_data,
                verbose=fit_config.monitor.verbose,
            )

        self.state = DeepEnsemblePosteriorStateRepository(
            ensemble_size=self.posterior_approximator.ensemble_size,
            checkpoint_dir=fit_config.checkpointer.save_checkpoint_dir
            if fit_config.checkpointer.dump_state is True
            else None,
        )
        status = []
        for i in range(self.posterior_approximator.ensemble_size):
            logging.info(
                f"Run {i+1} out of {self.posterior_approximator.ensemble_size}."
            )
            state, _status = _fit(i)
            self.state.put(
                state=state, i=i, keep=fit_config.checkpointer.keep_top_n_checkpoints
            )
            status.append(_status)
        logging.info("Fit completed.")
        return status

    def sample(self, rng: Optional[PRNGKeyArray] = None, **kwargs) -> JointState:
        # FIXME: Bug issue in JAX: https://github.com/google/jax/issues/13098
        # if rng is None:
        #     rng = self.rng.get()
        # state = pure_callback(
        #     lambda j: self.state.get(i=j),
        #     self.state.get(i=0),
        #     random.choice(rng, self.posterior_approximator.ensemble_size),
        # )
        state = self.state.get(
            i=np.random.choice(self.posterior_approximator.ensemble_size)
        )
        return JointState(
            params=state.params,
            mutable=state.mutable,
            calib_params=state.calib_params,
            calib_mutable=state.calib_mutable,
        )

    def load_state(self, checkpoint_dir: Path) -> None:
        try:
            self.restore_checkpoint(os.path.join(checkpoint_dir, "0"))
        except ValueError:
            raise ValueError(
                f"No checkpoint was found in `checkpoint_dir={checkpoint_dir}`."
            )
        self.state = DeepEnsemblePosteriorStateRepository(
            ensemble_size=self.posterior_approximator.ensemble_size,
            checkpoint_dir=checkpoint_dir,
        )

    def save_state(self, checkpoint_dir: Path, keep_top_n_checkpoints: int = 1) -> None:
        for i in range(self.posterior_approximator.ensemble_size):
            self.state.put(state=self.state.get(i), i=i, keep=keep_top_n_checkpoints)



from typing import Optional

import jax.numpy as jnp
import jax.scipy as jsp
from jax import jit, vmap
from jax._src.prng import PRNGKeyArray

from fortuna.data.loader import InputsLoader
from fortuna.prob_model.posterior.base import Posterior
from fortuna.prob_model.predictive.base import Predictive


class ClassificationPredictive(Predictive):
    def __init__(self, posterior: Posterior):
        """
        Classification predictive distribution class.

        Parameters
        ----------
        posterior : Posterior
             A posterior distribution object.
        """
        super().__init__(posterior)

    def mean(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive mean of the one-hot encoded target variable, that is

        .. math::
            \mathbb{E}_{\tilde{Y}|x, \mathcal{D}}[\tilde{Y}],

        where:
         - :math:`x` is an observed input variable;
         - :math:`\tilde{Y}` is a one-hot encoded random target variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`W` denotes the random model parameters.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        rng: Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive mean for each input.
        """
        return super().mean(inputs_loader, n_posterior_samples, rng, distribute)

    def mode(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        means: Optional[jnp.ndarray] = None,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        if means is None:
            means = self.mean(
                inputs_loader=inputs_loader,
                n_posterior_samples=n_posterior_samples,
                rng=rng,
                distribute=distribute,
            )
        return jnp.argmax(means, -1)

    def aleatoric_variance(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive aleatoric variance of the one-hot encoded target variable, that is

        .. math::
            \text{Var}_{W|\mathcal{D}}[\mathbb{E}_{\tilde{Y}|W, x}[\tilde{Y}]],

        where:
         - :math:`x` is an observed input variable;
         - :math:`\tilde{Y}` is a one-hot encoded random target variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`W` denotes the random model parameters.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive aleatoric variance for each input.
        """
        return super().aleatoric_variance(
            inputs_loader, n_posterior_samples, rng, distribute
        )

    def epistemic_variance(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
        **kwargs
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive epistemic variance of the one-hot encoded target variable, that is

        .. math::
            \mathbb{E}_{W|D}[\text{Var}_{\tilde{Y}|W, x}[\tilde{Y}]],

        where:
         - :math:`x` is an observed input variable;
         - :math:`\tilde{Y}` is a one-hot encoded random target variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`W` denotes the random model parameters.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive epistemic variance for each input.
        """
        return super().epistemic_variance(
            inputs_loader, n_posterior_samples, rng, distribute
        )

    def variance(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        aleatoric_variances: Optional[jnp.ndarray] = None,
        epistemic_variances: Optional[jnp.ndarray] = None,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive variance of the one-hot encoded target variable, that is

        .. math::
            \text{Var}_{\tilde{Y}|x, D}[\tilde{Y}],

        where:
         - :math:`x` is an observed input variable;
         - :math:`\tilde{Y}` is a one-hot encoded random target variable;
         - :math:`\mathcal{D}` is the observed training data set.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        aleatoric_variances: Optional[jnp.ndarray]
            An estimate of the aleatoric predictive variance.
        epistemic_variances: Optional[jnp.ndarray]
            An estimate of the epistemic predictive variance.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive variance for each input.
        """
        return super().variance(
            inputs_loader,
            n_posterior_samples,
            aleatoric_variances,
            epistemic_variances,
            rng,
            distribute,
        )

    def std(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        variances: Optional[jnp.ndarray] = None,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive standard deviation of the one-hot encoded target variable, that is

        .. math::
            \sqrt{\text{Var}_{\tilde{Y}|x, D}[\tilde{Y}]},

        where:
         - :math:`x` is an observed input variable;
         - :math:`\tilde{Y}` is a one-hot encoded random target variable;
         - :math:`\mathcal{D}` is the observed training data set.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        variances: Optional[jnp.ndarray]
            An estimate of the predictive variance.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive standard deviation for each input.
        """
        return super().std(
            inputs_loader, n_posterior_samples, variances, rng, distribute
        )

    def aleatoric_entropy(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive aleatoric entropy, that is

        .. math::
            -\mathbb{E}_{W|\mathcal{D}}[\mathbb{E}_{Y|W, x}[\log p(Y|W, x)]],

        where:
         - :math:`x` is an observed input variable;
         - :math:`Y` is a random target variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`W` denotes the random model parameters.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive aleatoric entropy for each input.
        """
        ensemble_outputs = self.sample_calibrated_outputs(
            inputs_loader=inputs_loader,
            n_output_samples=n_posterior_samples,
            rng=rng,
            distribute=distribute,
        )
        n_classes = ensemble_outputs.shape[-1]

        @jit
        def _entropy_term(i: int):
            targets = i * jnp.ones(ensemble_outputs.shape[1])

            def _log_lik_fun(outputs):
                return self.likelihood.prob_output_layer.log_prob(outputs, targets)

            log_liks = vmap(_log_lik_fun)(ensemble_outputs)
            return jnp.mean(jnp.exp(log_liks) * log_liks, 0)

        return -jnp.sum(vmap(_entropy_term)(jnp.arange(n_classes)), 0)

    def epistemic_entropy(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive epistemic entropy, that is

        .. math::
            -\mathbb{E}_{Y|x, \mathcal{D}}[\log p(Y|x, \mathcal{D})] +
            \mathbb{E}_{W|\mathcal{D}}[\mathbb{E}_{Y|W, x}[\log p(Y|W, x)]],

        where:
         - :math:`x` is an observed input variable;
         - :math:`Y` is a random target variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`W` denotes the random model parameters.

         Note that the epistemic entropy above is defined as the difference between the predictive entropy and the
         aleatoric predictive entropy.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive epistemic entropy for each input.
        """
        ensemble_outputs = self.sample_calibrated_outputs(
            inputs_loader=inputs_loader,
            n_output_samples=n_posterior_samples,
            rng=rng,
            distribute=distribute,
        )
        n_classes = ensemble_outputs.shape[-1]

        @jit
        def _entropy_term(i: int):
            targets = i * jnp.ones(ensemble_outputs.shape[1])

            def _log_lik_fun(outputs):
                return self.likelihood.prob_output_layer.log_prob(outputs, targets)

            log_liks = vmap(_log_lik_fun)(ensemble_outputs)
            log_preds = jsp.special.logsumexp(log_liks, 0) - jnp.log(
                n_posterior_samples
            )
            return jnp.exp(log_preds) * log_preds - jnp.mean(
                jnp.exp(log_liks) * log_liks, 0
            )

        return -jnp.sum(vmap(_entropy_term)(jnp.arange(n_classes)), 0)

    def entropy(
        self,
        inputs_loader: InputsLoader,
        n_posterior_samples: int = 30,
        rng: Optional[PRNGKeyArray] = None,
        distribute: bool = True,
    ) -> jnp.ndarray:
        r"""
        Estimate the predictive entropy, that is

        .. math::
            -\mathbb{E}_{Y|x, \mathcal{D}}[\log p(Y|x, \mathcal{D})],

        where:
         - :math:`x` is an observed input variable;
         - :math:`Y` is a random target variable;
         - :math:`\mathcal{D}` is the observed training data set;
         - :math:`W` denotes the random model parameters.

        Parameters
        ----------
        inputs_loader : InputsLoader
            A loader of input data points.
        n_posterior_samples : int
            Number of samples to draw from the posterior distribution for each input.
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        distribute: bool
            Whether to distribute computation over multiple devices, if available.

        Returns
        -------
        jnp.ndarray
            An estimate of the predictive entropy for each input.
        """
        ensemble_outputs = self.sample_calibrated_outputs(
            inputs_loader=inputs_loader,
            n_output_samples=n_posterior_samples,
            rng=rng,
            distribute=distribute,
        )
        n_classes = ensemble_outputs.shape[-1]

        @jit
        def _entropy_term(i: int):
            targets = i * jnp.ones(ensemble_outputs.shape[1])

            def _log_lik_fun(outputs):
                return self.likelihood.prob_output_layer.log_prob(outputs, targets)

            log_liks = vmap(_log_lik_fun)(ensemble_outputs)
            log_preds = jsp.special.logsumexp(log_liks, 0) - jnp.log(
                n_posterior_samples
            )
            return jnp.exp(log_preds) * log_preds

        return -jnp.sum(vmap(_entropy_term)(jnp.arange(n_classes)), 0)



from __future__ import annotations

from typing import Callable, Iterable, Optional, Union

import jax
import numpy as np
from flax import jax_utils
from jax.tree_util import tree_map

from fortuna.typing import Array, Batch


class DataLoader:
    def __init__(
        self,
        data_loader: Union[
            FromIterableToDataLoader,
            FromCallableIterableToDataLoader,
            FromArrayDataToDataLoader,
            FromTensorFlowDataLoaderToDataLoader,
            FromTorchDataLoaderToDataLoader,
            ChoppedDataLoader
        ],
    ):
        """
        A data loader class. Each batch is a tuple of input and target variables, respectively. Both inputs and targets
        are arrays, with different data points over the first dimension.

        Parameters
        ----------
        data_loader : Union[FromIterableToDataLoader, FromCallableIterableToDataLoader, FromArrayDataToDataLoader,
        FromTensorFlowDataLoaderToDataLoader, FromTorchDataLoaderToDataLoader]
            A data loader.
        """
        self._data_loader = data_loader

    def __iter__(self):
        yield from self._data_loader()

    @classmethod
    def from_array_data(
        cls,
        data: Batch,
        batch_size: Optional[int] = None,
        shuffle: bool = False,
        prefetch: bool = False,
    ) -> DataLoader:
        """
        Build a :class:`~fortuna.data.loader.DataLoader` object from a tuple of arrays of input and target variables,
        respectively.

        Parameters
        ----------
        data: Batch
            Input and target arrays of data.
        batch_size: Optional[int]
            The batch size. If not given, the data will not be batched.
        shuffle: bool
            Whether the data loader should shuffle at every call.
        prefetch: bool
            Whether to prefetch the next batch.

        Returns
        -------
        DataLoader
            A data loader built out of the tuple of arrays.
        """
        return cls(
            data_loader=FromArrayDataToDataLoader(
                data, batch_size=batch_size, shuffle=shuffle, prefetch=prefetch
            )
        )

    @classmethod
    def from_callable_iterable(cls, fun: Callable[[], Iterable[Batch],],) -> DataLoader:
        """
        Transform a callable iterable into a :class:`~fortuna.data.loader.DataLoader` object.

        Parameters
        ----------
        fun: Callable[[], Iterable[Batch]]
            A callable iterable of tuples of input and target arrays.

        Returns
        -------
        DataLoader
            A data loader object.
        """
        return cls(data_loader=FromCallableIterableToDataLoader(fun))

    @classmethod
    def from_iterable(cls, iterable: Iterable[Batch],) -> DataLoader:
        """
        Transform an iterable into a :class:`~fortuna.data.loader.DataLoader` object.

        Parameters
        ----------
        iterable: Iterable[Batch]
            An iterable of tuples of input and target arrays.

        Returns
        -------
        DataLoader
            A data loader object.
        """
        return cls(data_loader=FromIterableToDataLoader(iterable))

    @classmethod
    def from_tensorflow_data_loader(cls, tf_data_loader) -> DataLoader:
        """
        Transform a TensorFlow data loader into a :class:`~fortuna.data.loader.DataLoader` object.

        Parameters
        ----------
        tf_data_loader
            A TensorFlow data loader where each batch is a tuple of input and target Tensors.

        Returns
        -------
        DataLoader
            A data loader object.
        """
        return cls(
            data_loader=FromTensorFlowDataLoaderToDataLoader(
                tf_data_loader=tf_data_loader
            )
        )

    @classmethod
    def from_torch_data_loader(cls, torch_data_loader) -> DataLoader:
        """
        Transform a PyTorch data loader into a :class:`~fortuna.data.loader.DataLoader` object.

        Parameters
        ----------
        torch_data_loader
            A PyTorch data loader where each batch is a tuple of input and target Tensors.

        Returns
        -------
        DataLoader
            A data loader object.
        """
        return cls(
            data_loader=FromTorchDataLoaderToDataLoader(
                torch_data_loader=torch_data_loader
            )
        )

    def to_array_data(self) -> Batch:
        """
        Reduce a data loader to a tuple of input and target arrays.

        Returns
        -------
        Batch
            Tuple of input and target arrays.
        """
        inputs, targets = [], []
        for batch_inputs, batch_targets in self._data_loader():
            inputs.append(batch_inputs)
            targets.append(batch_targets)
        return np.concatenate(inputs, 0), np.concatenate(targets, 0)

    def to_array_inputs(self) -> Array:
        """
        Reduce a data loader to an array of target data.

        Returns
        -------
        Array
            Array of input data.
        """
        inputs = []
        for batch_inputs, batch_targets in self._data_loader():
            inputs.append(batch_inputs)
        return np.concatenate(inputs, 0)

    def to_array_targets(self) -> Array:
        """
        Reduce a data loader to an array of target data.

        Returns
        -------
        Array
            Array of input data.
        """
        targets = []
        for batch_inputs, batch_targets in self._data_loader():
            targets.append(batch_targets)
        return np.concatenate(targets, 0)

    def to_inputs_loader(self) -> InputsLoader:
        """
        Reduce a data loader to an inputs loader.

        Returns
        -------
        InputsLoader
            The inputs loader derived from the data loader.
        """
        return InputsLoader.from_data_loader(DataLoader(data_loader=self._data_loader))

    def to_targets_loader(self) -> TargetsLoader:
        """
        Reduce a data loader to a targets loader.

        Returns
        -------
        TargetsLoader
            The targets loader derived from the data loader.
        """
        return TargetsLoader.from_data_loader(DataLoader(data_loader=self._data_loader))

    @classmethod
    def chop(cls, data_loader: DataLoader, divisor: int) -> DataLoader:
        """
        Chop the last part of each batch of the data loader, to make sure the number od data points per batch divides
        `divisor`.

        Parameters
        ----------
        data_loader : DataLoader
            A data loader
        divisor : int
            Number of data points that each batched must divide.

        Returns
        -------
        DataLoader
            A data loader with chopped batches.
        """
        return cls(data_loader=ChoppedDataLoader(data_loader=data_loader, divisor=divisor))


class InputsLoader:
    def __init__(
        self,
        inputs_loader: Union[
            FromArrayInputsToInputsLoader,
            FromDataLoaderToInputsLoader,
            FromCallableIterableToInputsLoader,
            FromIterableToInputsLoader,
            ChoppedInputsLoader
        ],
    ):
        """
        An inputs loader class. Each batch is an array of inputs, with different data points over the first dimension.

        Parameters
        ----------
        inputs_loader : Union[FromArrayInputsToInputsLoader, FromDataLoaderToInputsLoader]
            An inputs loader.
        """
        self._inputs_loader = inputs_loader

    def __iter__(self):
        yield from self._inputs_loader()

    @classmethod
    def from_data_loader(cls, data_loader: DataLoader) -> InputsLoader:
        """
        Reduce a data loader to an inputs loader.

        Parameters
        ----------
        data_loader : DataLoader
            A data loader.

        Returns
        -------
        InputsLoader
            An inputs loader.
        """
        return cls(inputs_loader=FromDataLoaderToInputsLoader(data_loader))

    @classmethod
    def from_array_inputs(
        cls,
        inputs: Array,
        batch_size: Optional[int] = None,
        shuffle: bool = False,
        prefetch: bool = False,
    ) -> InputsLoader:
        """
        Build a :class:`~fortuna.data.loader.InputsLoader` object from an array of input data.

        Parameters
        ----------
        inputs: Array
            Input array of data.
        batch_size: Optional[int]
            The batch size. If not given, the inputs will not be batched.
        shuffle: bool
            Whether the inputs loader should shuffle at every call.
        prefetch: bool
            Whether to prefetch the next batch.

        Returns
        -------
        InputsLoader
            An inputs loader built out of the array of inputs.
        """
        return cls(
            inputs_loader=FromArrayInputsToInputsLoader(
                inputs, batch_size=batch_size, shuffle=shuffle, prefetch=prefetch
            )
        )

    def to_array_inputs(self) -> Array:
        """
        Reduce an inputs loader to an array of inputs.

        Returns
        -------
        Array
            Array of input data.
        """
        inputs = []
        for batch_inputs in self._inputs_loader():
            inputs.append(batch_inputs)
        return np.concatenate(inputs, 0)

    @classmethod
    def from_callable_iterable(
        cls, fun: Callable[[], Iterable[Array]],
    ) -> InputsLoader:
        """
        Transform a callable iterable into a :class:`~fortuna.data.loader.InputsLoader` object.

        Parameters
        ----------
        fun: Callable[[], Iterable[Array]]
            A callable iterable of input arrays.

        Returns
        -------
        InputsLoader
            An inputs loader object.
        """
        return cls(inputs_loader=FromCallableIterableToInputsLoader(fun))

    @classmethod
    def from_iterable(cls, iterable: Iterable[Array],) -> InputsLoader:
        """
        Transform an iterable into a :class:`~fortuna.data.loader.InputsLoader` object.

        Parameters
        ----------
        iterable: Iterable[Array]
            An iterable of input arrays.

        Returns
        -------
        InputsLoader
            An inputs loader object.
        """
        return cls(inputs_loader=FromIterableToInputsLoader(iterable))

    @classmethod
    def chop(cls, inputs_loader: InputsLoader, divisor: int) -> InputsLoader:
        """
        Chop the last part of each batch of the inputs loader, to make sure the number od data points per batch divides
        `divisor`.

        Parameters
        ----------
        inputs_loader : InputsLoader
            An inputs loader.
        divisor : int
            Number of data points that each batched must divide.

        Returns
        -------
        InputsLoader
            An inputs loader with chopped batches.
        """
        return cls(inputs_loader=ChoppedInputsLoader(inputs_loader=inputs_loader, divisor=divisor))


class TargetsLoader:
    def __init__(
        self,
        targets_loader: Union[
            FromArrayTargetsToTargetsLoader,
            FromDataLoaderToTargetsLoader,
            FromCallableIterableToTargetsLoader,
            FromIterableToTargetsLoader,
            ChoppedTargetsLoader
        ],
    ):
        """
        A targets loader class. Each batch is an array of targets, with different data points over the first dimension.

        Parameters
        ----------
        targets_loader : Union[FromArrayTargetsToTargetsLoader, FromDataLoaderToTargetsLoader]
            A targets loader.
        """
        self._targets_loader = targets_loader

    def __iter__(self):
        yield from self._targets_loader()

    @classmethod
    def from_data_loader(cls, data_loader: DataLoader) -> TargetsLoader:
        """
        Reduce a data loader to a targets loader.

        Parameters
        ----------
        data_loader : DataLoader
            A data loader.

        Returns
        -------
        TargetsLoader
            A targets loader.
        """
        return cls(targets_loader=FromDataLoaderToTargetsLoader(data_loader))

    @classmethod
    def from_array_targets(
        cls,
        targets: Array,
        batch_size: Optional[int] = None,
        shuffle: bool = False,
        prefetch: bool = False,
    ) -> TargetsLoader:
        """
        Build a :class:`~fortuna.data.loader.TargetsLoader` object from an array of target data.

        Parameters
        ----------
        targets: Array
            Target array of data.
        batch_size: Optional[int]
            The batch size. If not given, the targets will not be batched.
        shuffle: bool
            Whether the target loader should shuffle at every call.
        prefetch: bool
            Whether to prefetch the next batch.

        Returns
        -------
        TargetsLoader
            A targets loader built out of the array of targets.
        """
        return cls(
            targets_loader=FromArrayTargetsToTargetsLoader(
                targets, batch_size=batch_size, shuffle=shuffle, prefetch=prefetch
            )
        )

    def to_array_targets(self) -> Array:
        """
        Reduce a targets loader to an array of targets.

        Returns
        -------
        Array
            Array of target data.
        """
        targets = []
        for batch_targets in self._targets_loader():
            targets.append(batch_targets)
        return np.concatenate(targets, 0)

    @classmethod
    def from_callable_iterable(
        cls, fun: Callable[[], Iterable[Array]],
    ) -> TargetsLoader:
        """
        Transform a callable iterable into a :class:`~fortuna.data.loader.TargetsLoader` object.

        Parameters
        ----------
        fun: Callable[[], Iterable[Array]]
            A callable iterable of target arrays.

        Returns
        -------
        TargetsLoader
            A targets loader object.
        """
        return cls(targets_loader=FromCallableIterableToTargetsLoader(fun))

    @classmethod
    def from_iterable(cls, iterable: Iterable[Array],) -> TargetsLoader:
        """
        Transform an iterable into a :class:`~fortuna.data.loader.TargetsLoader` object.

        Parameters
        ----------
        iterable: Iterable[Array]
            An iterable of target arrays.

        Returns
        -------
        TargetsLoader
            A targets loader object.
        """
        return cls(targets_loader=FromIterableToTargetsLoader(iterable))

    @classmethod
    def chop(cls, targets_loader: TargetsLoader, divisor: int) -> TargetsLoader:
        """
        Chop the last part of each batch of the targets loader, to make sure the number od data points per batch divides
        `divisor`.

        Parameters
        ----------
        targets_loader : TargetsLoader
            A targets loader.
        divisor : int
            Number of data points that each batched must divide.

        Returns
        -------
        InputsLoader
            A targets loader with chopped batches.
        """
        return cls(targets_loader=ChoppedTargetsLoader(targets_loader=targets_loader, divisor=divisor))


class FromDataLoaderToArrayData:
    def __init__(self, data_loader: DataLoader):
        self._data_loader = data_loader

    def __call__(self, *args, **kwargs):
        data = []
        for batch in self._data_loader:
            data.append(batch)
        return np.concatenate(data, 0)


class FromDataLoaderToInputsTargetsLoaders:
    def __init__(self, data_loader: DataLoader):
        self._data_loader = data_loader

    def __call__(self, *args, **kwargs):
        for x_batch, y_batch in self._data_loader:
            yield x_batch, y_batch


class FromArrayDataToDataLoader:
    def __init__(
        self,
        data: Batch,
        batch_size: Optional[int] = None,
        shuffle: bool = False,
        prefetch: bool = False,
    ):
        self._data = data
        self._batch_size = batch_size
        self._shuflle = shuffle
        self._prefetch = prefetch

    def __call__(self, *args, **kwargs):
        if self._shuflle:
            perm = np.random.choice(
                self._data[0].shape[0], self._data[0].shape[0], replace=False
            )
        if self._batch_size is None:
            yield self._data
        else:
            x_batches = np.split(
                self._data[0][perm] if self._shuflle else self._data[0],
                np.arange(self._batch_size, self._data[0].shape[0], self._batch_size),
                axis=0,
            )
            y_batches = np.split(
                self._data[1][perm] if self._shuflle else self._data[1],
                np.arange(self._batch_size, self._data[1].shape[0], self._batch_size),
                axis=0,
            )

            def make_gen():
                for x_batch, y_batch in zip(x_batches, y_batches):
                    yield x_batch, y_batch

            yield from PrefetchedGenerator(make_gen()) if self._prefetch else make_gen()


class FromCallableIterableToDataLoader:
    def __init__(
        self, fun: Callable[[], Iterable[Batch],],
    ):
        self._fun = fun

    def __call__(self, *args, **kwargs):
        return self._fun()


class FromCallableIterableToInputsLoader:
    def __init__(
        self, fun: Callable[[], Iterable[Array]],
    ):
        self._fun = fun

    def __call__(self, *args, **kwargs):
        return self._fun()


class FromCallableIterableToTargetsLoader:
    def __init__(
        self, fun: Callable[[], Iterable[Array]],
    ):
        self._fun = fun

    def __call__(self, *args, **kwargs):
        return self._fun()


class FromIterableToDataLoader:
    def __init__(
        self, batched_data: Iterable[Batch],
    ):
        self._batched_data = batched_data

    def __call__(self, *args, **kwargs):
        return self._batched_data


class FromIterableToInputsLoader:
    def __init__(
        self, batched_inputs: Iterable[Array],
    ):
        self._batched_inputs = batched_inputs

    def __call__(self, *args, **kwargs):
        return self._batched_inputs


class FromIterableToTargetsLoader:
    def __init__(
        self, batched_targets: Iterable[Array],
    ):
        self._batched_targets = batched_targets

    def __call__(self, *args, **kwargs):
        return self._batched_targets


class FromTensorFlowDataLoaderToDataLoader:
    def __init__(self, tf_data_loader):
        self._tf_data_loader = tf_data_loader

    def __call__(self, *args, **kwargs):
        for batch_inputs, batch_targets in self._tf_data_loader:
            yield batch_inputs.numpy(), batch_targets.numpy()


class FromTorchDataLoaderToDataLoader:
    def __init__(self, torch_data_loader):
        self._torch_data_loader = torch_data_loader

    def __call__(self, *args, **kwargs):
        for batch_inputs, batch_targets in self._torch_data_loader:
            yield batch_inputs.numpy(), batch_targets.numpy()


class FromDataLoaderToArrayInputs:
    def __init__(self, data_loader: DataLoader):
        self._data_loader = data_loader

    def __call__(self, *args, **kwargs):
        inputs = []
        for batch_inputs, batch_targets in self._data_loader:
            inputs.append(batch_inputs)
        return np.concatenate(inputs, 0)


class FromDataLoaderToArrayTargets:
    def __init__(self, data_loader: DataLoader):
        self._data_loader = data_loader

    def __call__(self, *args, **kwargs):
        targets = []
        for batch_inputs, batch_targets in self._data_loader:
            targets.append(batch_targets)
        return np.concatenate(targets, 0)


class FromInputsLoaderToArrayInputs:
    def __init__(self, inputs_loader: InputsLoader):
        self._inputs_loader = inputs_loader

    def __call__(self, *args, **kwargs):
        inputs = []
        for batch_inputs in self._inputs_loader:
            inputs.append(batch_inputs)
        return np.concatenate(inputs, 0)


class FromArrayInputsToInputsLoader:
    def __init__(
        self,
        inputs: Array,
        batch_size: Optional[int] = None,
        shuffle: bool = False,
        prefetch: bool = False,
    ):
        self._inputs = inputs
        self._batch_size = batch_size
        self._shuflle = shuffle
        self._prefetch = prefetch

    def __call__(self, *args, **kwargs):
        if self._shuflle:
            perm = np.random.choice(
                self._inputs.shape[0], self._inputs.shape[0], replace=False
            )
        if self._batch_size is None:
            yield self._inputs
        else:
            x_batches = np.split(
                self._inputs[perm] if self._shuflle else self._inputs,
                np.arange(self._batch_size, self._inputs.shape[0], self._batch_size),
                axis=0,
            )

            def make_gen():
                for x_batch in x_batches:
                    yield x_batch

            yield from PrefetchedGenerator(make_gen()) if self._prefetch else make_gen()


class FromDataLoaderToInputsLoader:
    def __init__(
        self, data_loader: DataLoader,
    ):
        self._data_loader = data_loader

    def __call__(self, *args, **kwargs):
        for inputs, targets in self._data_loader:
            yield inputs


class FromDataLoaderToTargetsLoader:
    def __init__(
        self, data_loader: DataLoader,
    ):
        self._data_loader = data_loader

    def __call__(self, *args, **kwargs):
        for inputs, targets in self._data_loader:
            yield targets


class FromTargetsLoaderToArrayTargets:
    def __init__(self, targets_loader: TargetsLoader):
        self._targets_loader = targets_loader

    def __call__(self, *args, **kwargs):
        targets = []
        for batch_targets in self._targets_loader:
            targets.append(batch_targets)
        return np.concatenate(targets, 0)


class FromArrayTargetsToTargetsLoader:
    def __init__(
        self,
        targets: Array,
        batch_size: Optional[int] = None,
        shuffle: bool = False,
        prefetch: bool = False,
    ):
        self._targets = targets
        self._batch_size = batch_size
        self._shuflle = shuffle
        self._prefetch = prefetch

    def __call__(self, *args, **kwargs):
        if self._shuflle:
            perm = np.random.choice(
                self._targets.shape[0], self._targets.shape[0], replace=False
            )
        if self._batch_size is None:
            yield self._targets
        else:
            x_batches = np.split(
                self._targets[perm] if self._shuflle else self._targets,
                np.arange(self._batch_size, self._targets.shape[0], self._batch_size),
                axis=0,
            )

            def make_gen():
                for x_batch in x_batches:
                    yield x_batch

            yield from PrefetchedGenerator(make_gen()) if self._prefetch else make_gen()


class PrefetchedGenerator:
    def __init__(self, generator):
        self._batch = generator.__next__()
        self._generator = generator
        self._ready = True

    def __iter__(self):
        return self

    def __next__(self):
        if not self._ready:
            self._prefetch()
        self._ready = False
        return self._batch

    def prefetch(self):
        if not self._ready:
            self._batch = self._generator.__next__()
            self._ready = True


class ChoppedDataLoader:
    def __init__(self, data_loader: DataLoader, divisor: int):
        self._data_loader = data_loader
        self._divisor = divisor

    def __call__(self, *args, **kwargs):
        for inputs, targets in self._data_loader:
            reminder = targets.shape[0] % self._divisor
            if reminder == 0:
                yield inputs, targets
            else:
                yield inputs[:-reminder], targets[:-reminder]


class ChoppedInputsLoader:
    def __init__(self, inputs_loader: InputsLoader, divisor: int):
        self._inputs_loader = inputs_loader
        self._divisor = divisor

    def __call__(self, *args, **kwargs):
        for inputs in self._inputs_loader:
            reminder = inputs.shape[0] % self._divisor
            if reminder == 0:
                yield inputs
            else:
                yield inputs[:-reminder]


class ChoppedTargetsLoader:
    def __init__(self, targets_loader: TargetsLoader, divisor: int):
        self._targets_loader = targets_loader
        self._divisor = divisor

    def __call__(self, *args, **kwargs):
        for targets in self._targets_loader:
            reminder = targets.shape[0] % self._divisor
            if reminder == 0:
                yield targets
            else:
                yield targets[:-reminder]


class DeviceDimensionAugmentedDataLoader:
    def __init__(self, data_loader: DataLoader):
        self._data_loader = data_loader

    @staticmethod
    def _reshape_batch(batch):
        n_devices = jax.local_device_count()
        if batch.shape[0] % n_devices != 0:
            raise ValueError(
                f"The size of all batches must be a multiple of {n_devices}, that is the number of "
                f"available devices. However, a batch with shape {batch.shape[0]} was found. "
                f"Please set an appropriate batch size."
            )
        return batch.reshape((n_devices, -1) + batch.shape[1:])

    def __iter__(self, *args, **kwargs):
        data_loader = map(
            lambda batch: tree_map(self._reshape_batch, batch), self._data_loader
        )
        data_loader = jax_utils.prefetch_to_device(data_loader, 2)
        yield from data_loader


class DeviceDimensionAugmentedInputsLoader:
    def __init__(self, inputs_loader: InputsLoader):
        self._inputs_loader = inputs_loader

    @staticmethod
    def _reshape_inputs(inputs):
        n_devices = jax.local_device_count()
        if inputs.shape[0] % n_devices != 0:
            raise ValueError(
                f"The size of all batches of inputs must be a multiple of {n_devices}, that is the number of "
                f"available devices. However, a batch of inputs with shape {inputs.shape[0]} was found. "
                f"Please set an appropriate batch size."
            )
        return inputs.reshape((n_devices, -1) + inputs.shape[1:])

    def __iter__(self, *args, **kwargs):
        inputs_loader = map(
            lambda inputs: self._reshape_inputs(inputs), self._inputs_loader
        )
        inputs_loader = jax_utils.prefetch_to_device(inputs_loader, 2)
        yield from inputs_loader
